# Generated by Django 3.2 on 2022-11-23 11:21

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('apis_entities', '0004_auto_20200722_1231'),
        ('dubletten_tool', '0006_delete_ampel'),
    ]

    operations = [
        migrations.AddField(
            model_name='group',
            name='vorfin',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL, to='apis_entities.person'),
        ),
    ]



from . import log

class ErrorLoggerMixin:
    """
    Helper class to be subclassed by all classes that process data and might throw an error.
    Logs the class, the type of error and after all processing is done, logs which 
    classes threw errors, and which passed.
    First logs an overview with the number of errors per class and errortype.
    Also logs a detailed list of all individual error messages per class and errortype.

    # todo: unsure how to reset this class, as the subclasses have each there own data dicts.
    """
    enum = None
    collected_errors = {}

    @classmethod
    def setup(cls):
        cls.enum = None
        cls.collected_errors = {}
        
    @classmethod
    def has_errors(cls):
        """
        Short helper to check if a class has errors or not.
        """
        if cls.errors.keys() or cls.enum_errors.keys():
            return True
        else:
            return False
    
    @classmethod
    def log_report(cls):
        """
        Logs if a class threw errors and if so, how many.
        """
        cls_name = cls.__name__
        has_errors = {}

        if not cls.has_errors():
            log(f"Report from {cls_name}: No Errors.")
        else: 
            log(f"{cls_name} Errors: ")
            for k, v in cls.errors.items():
                log(f"{k}: {len(v)}")
                has_errors.update({k:v})

            if cls.enum:
                log(f"{cls.enum.__name__} errors: ")
                for k, v in cls.enum_errors.items():
                    log(f"'{k}'-count: {len(v)}")
                    has_errors.update({k:v})
        
        
        if has_errors.keys():  
            cls.collected_errors.update({cls:has_errors})
 
    @classmethod               
    def log_details(cls):
        """
        Logs the individual error messages per class and errortype.
        """
        for c, errors in cls.collected_errors.items():
            for k,v in errors.items():
                log(f"Detailed Errorlog for {c} [{k}]")
                for idx, el in enumerate(v):
                    log(f"\tErr_{idx}: {el}")

        
    


from django.test import TestCase



class RelationWriterTest(TestCase):

    def setup(self, *args, **kwargs):
        pass

    def test_class_setup(self, *args, **kwargs):
        pass

    def tets_class_reset(self, *args, **kwargs):
        pass

    



from .error_logger_mixin_test import ErrorLoggerMixinTest
from .merge_group_test import MergeGroupTest
from .person_helper_test import PersonHelperTest
from .pphelper_test import PPHelperTest
from .person_type_test import PersonTypeTest
from .ppkind_test import PPKindTest
from .relation_writer_test import RelationWriterTest
from .single_merge_test import SingelMergeTest
from .temp_rel_test import TempRelTest



from . import log, defaultdict, Collection, LabelType, PersonPersonRelation, logger, PersonProxy, Person, AbstractRelation, PersonPerson, Label


class MergeGroup:
    """
    Functions defined here handle the merging of groups into a vorfinalen eintrag (MergeGroup-Class) 
    and the writing of personperson relations in existing vorfinale Einträge from the related Group.

    This is the original Merging script, updated to better process the different kinds of Person involved in 
    PerPer Relations. 

    - also processes Uris now
    - also processes titles now

    !!!!!!!!!
    IMPORTANT: 
    IF RUN IN NOTEBOOK: the main method to writing person person rels is extracted into the RelationWriter class
    and its method is used in the script, not the one defined here !
    !!!!!!!!!

    These functions are used: 
    - in the management command to create the inital merges (merge_groups command)
    - in merge_views.py, in the remergegroupview, that can be called in the dubletten_tool frontend section. The template for the button is member_list.html, the callback function resides in tool_page.html
    """

    merged_col, c = Collection.objects.get_or_create(name="Vorfinale Einträge")
    rt_vorfin, c = PersonPersonRelation.objects.get_or_create(
        name="data merged from", name_reverse="merged into"
    )
    lt_group, c = LabelType.objects.get_or_create(
        name="Result of deduplication Group")
    lt_first_name_alt, c = LabelType.objects.get_or_create(
        name="alternativer Vorname")
    lt_name_alt, c = LabelType.objects.get_or_create(
        name="alternativer Nachname")
    lt_sd_alt, c = LabelType.objects.get_or_create(
        name="alternatives Geburtsdatum")
    lt_ed_alt, c = LabelType.objects.get_or_create(
        name="alternatives Sterbedatum")
    #group_map = {} unused, therefore removed

    rels = [
        "PersonInstitution",
        "PersonPlace",
        "PersonEvent",
        "PersonWork",
        "PersonPerson",
    ]  # Handle seperately:, "PersonPerson"]

    def __init__(self, group, create=False):
        self.group = group
        self.members = [m.person for m in group.members.all()]
        self.meta = {
            "name": self.members[0].name,
            "first_name": self.members[0].first_name,
            "gender": self.group.gender,
        }
        self.create = create
        self.titles = []
        # self.additional_labels = []

    def process_names(self):
        for m in self.members[1:]:
            self.labels[MergeGroup.lt_first_name_alt].append(
                {"label": m.first_name})
            self.labels[MergeGroup.lt_name_alt].append({"label": m.name})

    def process_birth_and_death(self):
        start_dates = set()
        end_dates = set()
        alt_start, alt_end = None, None
        for m in self.members:
            if m.start_date_written:
                start_dates.add(m.start_date_written)
            if m.end_date_written:
                end_dates.add(m.end_date_written)

        start_dates = list(start_dates)
        end_dates = list(end_dates)

        if start_dates:
            if len(start_dates) == 1:
                start_date = start_dates[0]
            else:
                start_date = start_dates[0]
                alt_start = start_dates[1:]

            #  start_date, *alt_start = [start_dates]

        else:
            start_date = None

        if end_dates:
            if len(end_dates) == 1:
                end_date = end_dates[0]
            else:
                end_date = end_dates[0]
                alt_end = end_dates[1:]
        else:
            end_date = None

        self.meta.update(
            {"start_date_written": start_date, "end_date_written": end_date}
        )

        # todo: create labels from alternative start and end date lists
        if alt_start:
            for sd in alt_start:
                self.labels[MergeGroup.lt_sd_alt].append({"label": sd})
        if alt_end:
            for ed in alt_end:
                self.labels[MergeGroup.lt_ed_alt].append({"label": ed})

            # self.additional_labels.append(label_type=MergeGroup.lt_ed_alt, label=ed)

    def process_labels(self):
        labels = defaultdict(list)
        for m in self.members:
            [
                labels[el.label_type].append(
                    {
                        "label": el.label.strip(),
                        "start_date_written": el.start_date_written,
                        "end_date_written": el.end_date_written,
                    }
                )
                for el in m.label_set.all()
            ]
        # print("labels is", labels, labels.items())

        return labels

    def process_personinstitution(self):
        pass

    def process_all_other_relations(self):
        # todo: alle relations deduplizieren
        for m in self.members:
            for r in MergeGroup.rels:
                # print("r is:", r)
                if r != "PersonPerson":
                    temp = getattr(m, r.lower() + "_set").values()
                    model = AbstractRelation.get_relation_class_of_name(r)
                    # print(temp)
                    for t in temp:
                        t.pop("related_person_id")
                        t.pop("tempentityclass_ptr_id")
                        t.pop("id")
                        t["related_person"] = self.person
                        # print(t)
                        model.objects.get_or_create(**t)
                elif r == "PersonPerson":
                    # for d in ["personA_set", "personB_set"]:
                    # tempA = m.personA_set.values()
                    # tempB = m.personB_set.values()
                    pass

                    # for t in tempA:
                    #     print(t)
                    #     t.pop("related_personA_id")
                    #     t.pop("tempentityclass_ptr_id")
                    #     t.pop("id")
                    #     t["related_personA"] = self.person
                    #     #print(t)
                    #     model.objects.get_or_create(**t)
                    # for t in tempB:
                    #     print(t)
                    #     t.pop("related_personB_id")
                    #     t.pop("tempentityclass_ptr_id")
                    #     t.pop("id")
                    #     t["related_personB"] = self.person
                    #     #print(t)
                    #     model.objects.get_or_create(**t)

    def process_titles(self):
        titels = set()
        for m in self.members:
            [titels.add(t) for t in m.title.all()]

        return titels

    def process_notes(self):
        notes = ""
        for m in self.members:
            if m.notes:
                notes += f"[{m.pk}: {m.notes}]\n"
        return notes

    def process_references(self):
        refs = ""
        for m in self.members:
            if m.references:
                refs += f"[{m.pk}: {m.references}]\n"
        return refs

    def process_uris(self):
        uris = set()
        for m in self.members:
            for uri in m.person.uri_set.all():
                if uri.domain != "apis default":
                    uris.add(uri)
        return uris

    def process_collections(self):
        # nicht nötig
        pass

    def run_process(self):

        self.labels = self.process_labels()
        self.process_names()
        self.process_birth_and_death()

        self.titles = self.process_titles()
        notes = self.process_notes()
        refs = self.process_references()
        self.meta.update({"notes": notes, "references": refs})
        per = Person.objects.create(**self.meta)
        per.name = per.name.strip() + " {vorfinal}"
        per.collection.add(MergeGroup.merged_col)
        [per.title.add(t) for t in self.titles]
        self.uris = self.process_uris()
        for uri in self.uris:
            per.uri_set.add(uri)

        per.save()
        # MergeGroup.group_map.update({self.group.id:per.id})
        self.group.vorfin = per
        self.group.save()
        print("merged group")
        # create Ampel
        # ampel = Ampel.objects.create(person=per, status="red", note=self.group.note)
        # ampel.save()

        # todo: handle start and end dates of labels as tuples
        # print(self.labels, self.labels.items())
        for key, vals in self.labels.items():
            for v in vals:
                try:
                    lab, created = Label.objects.get_or_create(
                        label_type=key, temp_entity=per, **v
                    )
                except Exception as e:
                    logger.exception(e)
                    continue
                # per.label_set.add(lab)
        for m in self.members:
            PersonPerson.objects.get_or_create(
                related_personA=per,
                related_personB=m,
                relation_type=MergeGroup.rt_vorfin,
            )

        Label.objects.get_or_create(
            label=f"{self.group.name} ({self.group.id})",
            label_type=MergeGroup.lt_group,
            temp_entity=per,
        )
        self.person = per
        self.process_all_other_relations()
        ###print("Notes:", notes, "refs:", refs)


class CountPerPer:
    single_count = 0
    member_count = 0


# def write_person_person_rels(g):
#     model = PersonPerson
#     vorfin = g.vorfin  # Person.objects.get(id=MergeGroup.group_map[g.id])
#     print("vorfin is", vorfin)
#     rt_vorfin, c = PersonPersonRelation.objects.get_or_create(
#         name="data merged from", name_reverse="merged into"
#     )

#     def get_dublette_or_vorfin(per):
#         print("called get_dublette_or_vorfin with per = ", per)
#         res = None
#         if PersonProxy.objects.filter(person=per).exists():
#             print("person existed")
#             proxy = PersonProxy.objects.get(person=per)

#             if proxy.status == "single":
#                 print("personproxy status was single")
#                 CountPerPer.single_count += 1
#                 print("person was single", per)
#                 res = per
#             else:
#                 print("person status was not single")

#                 gs = proxy.group_set.all()
#                 if len(gs) > 1:
#                     print(
#                         f"personproxy {proxy} {proxy.id} was member in {gs} - this is {len(gs)}"
#                     )
#                 res = gs[
#                     0
#                 ].vorfin  # Person.objects.get(id=MergeGroup.group_map[gs[0].id])

#                 if not res:
#                     res = per
#                     print("person was not single but also not found in group")

#                 else:
#                     CountPerPer.member_count += 1
#         else:
#             print("person was no PersonProxy", per)
#             res = per
#         print("res is: ", res)
#         return res

#     for m in g.members.all():
#         ppA = PersonPerson.objects.filter(related_personA=m.person)
#         ppB = PersonPerson.objects.filter(related_personB=m.person)
#         print("member is", m)
#         print(f"ppA {ppA}, ppB {ppB}")
#         if ppA:
#             for el in ppA:
#                 print("in ppA, el is: ", el)
#                 if el.relation_type != rt_vorfin:
#                     perB = get_dublette_or_vorfin(el.related_personB)
#                     print("perB is: ", perB)
#                     print("related_personA is ", vorfin)
#                     temp = {
#                         "relation_type": el.relation_type,
#                         "start_date_written": el.start_date_written,
#                         "end_date_written": el.end_date_written,
#                         "related_personA": vorfin,
#                         "related_personB": perB,
#                     }
#                     model.objects.get_or_create(**temp)
#         if ppB:
#             for el in ppB:
#                 print("in ppB, el is: ", el)
#                 if el.relation_type != rt_vorfin:
#                     perA = get_dublette_or_vorfin(el.related_personA)
#                     print("perA is: ", perA)
#                     print("realted_personB is: ", vorfin)
#                     temp = {
#                         "relation_type": el.relation_type,
#                         "start_date_written": el.start_date_written,
#                         "end_date_written": el.end_date_written,
#                         "related_personA": perA,
#                         "related_personB": vorfin,
#                     }
#                     model.objects.get_or_create(**temp)





from . import get_ampel, Person, PersonHelper, Group, AbstractEntity, logger, PersonPerson, ErrorLoggerMixin, get_group, defaultdict, Counter, AmpelTemp, AbstractRelation, log, PersonType, PPKind, pd


class TempRel(ErrorLoggerMixin):
    """
    Helper Class that stores, classifies, and re_writes all relations that might have been 
    created after the initial merge in Vorfins with red status. 

    The idea is: 

    - it stores all data from all relations (all relation models) that red vorfins participate in.
    - it checks related_entityA and related_entityB
    - if an related entity is a VORFIN, it stores the group
    - if an related entity is a dublette, it also stores the group
    - on re-write, it fetches the newly created vorfin for each group
    - and it checks if the relation is already present (get_or_create)
    - it also stores logging data and outputs a final report to the logfile
    - it also writes the re-created relations to an excel file, for manual checks.
    """

    @classmethod
    def setup(cls):
        cls._instances = []
        #_instancesA = defaultdict(list)
        #_instancesB = defaultdict(list)
        cls._created_counter = 0
        cls.created = defaultdict(list)
        cls.existed = []
        cls.created_by_model = Counter()
        cls.existed_by_model = Counter()

        cls.errors = defaultdict(list)
        cls.enum_errors = defaultdict(list)

        cls.changed_edited_vorfins = []

    @classmethod
    def reset(cls):
        cls.setup()

    def __init__(self, rel):
        """
        Stores the data for a given relation as a TempRel - Instance. 
        All instances are stored in the TempRel classes' _instances list. 
        Processing of related_entityA and related_entityB happens here.

        All fields that are to be preserved need to be stored here. 
        """
        self.model = self.get_rel_model(rel)
        ent_a, ent_b = self.get_related_entities(rel)

        # related_entityA
        self.a = TempRel.save_group_or_entity(ent_a)
        if not self.a:
            msg = f"Returned None for ent_a: {ent_a}, from rel {rel}."
            TempRel.errors["Returned None Error"].append(msg)

        # related_entityB
        self.b = TempRel.save_group_or_entity(ent_b)
        if not self.b:
            msg = f"Returned None for ent_b: {ent_b}, from rel {rel}."
            TempRel.errors["Returned None Error"].append(msg)

        # relation_type
        self.rt = rel.relation_type

        # start and end date written
        self.start_date_written = rel.start_date_written
        self.end_date_written = rel.end_date_written

        # if relation hat an active ampel status, save the status, otherwise none (used to check in re-create-part)
        if AmpelTemp.objects.filter(temp_entity=rel):
            self.ampel = rel.ampel.status
        else:
            self.ampel = None

        # store all instances in a list. they are never accessed individually, but bulk-processed
        TempRel._instances.append(self)

    def get_rel_model(self, rel):
        """
        Helper to get the model of a relation.
        """
        model_name = rel.__class__.__name__
        model = AbstractRelation.get_relation_class_of_name(model_name)
        return model

    @property
    def a_name(self):
        """
        managed property to get the name of the a-side field.
        """
        return self.model.get_related_entity_field_nameA()

    @property
    def b_name(self):
        """
        managed property to get the name of the b-side field.
        """
        return self.model.get_related_entity_field_nameB()

    def get_related_entities(self, rel):
        """
        helper function to get entityA and entityB instances from relation.
        """

        ent_a = getattr(rel, self.a_name)
        ent_b = getattr(rel, self.b_name)

        if not ent_a:
            msg = f"None Error: ent_a: {ent_a}, rel: {rel}."
            TempRel.errors["Related Entity was none Error"].append(msg)
            log(msg)
        if not ent_b:
            msg = f"None Error: ent_b: {ent_b}, rel: {rel}."
            TempRel.errors["Related Entity was none Error"].append(msg)
            log(msg)

        return ent_a, ent_b

    @classmethod
    def is_vorfin(cls, ent):
        """
        Returns true if entity is a person and in vorfins.
        """
        if ent in PersonHelper._vorfins:
            return True
        else:
            return False

    @classmethod
    def vorfin_not_red_check(cls, vorfin):
        """
        expects a vorfin and returns true of the ampel of the vorfin is green or yellow.
        """
        status = get_ampel(vorfin)
        if status in ["yellow", "green"]:
            return True
        else:
            return False

    def get_dict_data(self):
        """
        helper function for serialising the stored data. 
        To allow dictionary destructuring in get_or_create.
        """

        temp = {
            "relation_type": self.rt,
            "start_date_written": self.start_date_written,
            "end_date_written": self.end_date_written,
        }

        # This part takes care of the diverging related_entity field names in Relation models.
        temp[self.a_name] = self.get_vorfin_or_entity(self.a)
        temp[self.b_name] = self.get_vorfin_or_entity(self.b)

        return temp

    @classmethod
    def save_group_or_entity(cls, ent):
        """
        Helper function that expects an entity and returns said entity, or, if it 
        is an instance of Person, checks if it is a VORFIN. if, so, returnes the group,
        instead of the entity.
        Also checks if it is a Dublette, if so, tries to get teh group and returns the group.
        """

        log(f"in save_group_or_person, ent is {ent}")
        if isinstance(ent, Person):
            kind = PersonType.classify(ent)
            log(f"Classified {ent} as {kind}")
            # if ent in PersonHelper._vorfins:
            if kind == PersonType.VORFIN:
                res = get_group(ent)
                log(f"Entity was Person and was in vorfins, returning: {res}")
                return res
            # elif ent in PersonHelper._dubletten:
            if kind == PersonType.DUBLETTE:
                log(f"kind of {ent} was Dublette, entered processing")
                # try:
                proxy = ent.personproxy
                if not proxy:
                    msg = f"No Proxy Error: person: {ent}. proxy was: {proxy}"
                    cls.errors["Dublette without proxy error"].append(msg)
                    log(msg)
                    return ent
                groupset = proxy.group_set.all()
                if len(groupset) > 1:
                    msg = f"Dublette {ent} in manually created PP-rels has multiple groups: {groupset}. Wrote to first group"
                    cls.errors["Dublette with multiple groups Error"].append(
                        msg)
                    log(msg)

                group = groupset[0]
                if not group:
                    msg = f"Dublette without group: {ent}, group: {group}"
                    cls.errors["Dublette without group Error"].append(msg)
                log(f"returning group: {group}")
                return group
            else:
                log(f"kind of person was: {kind}, returning person {ent}")
                return ent

                # except Exception as e:
                #cls.errors["Raised Exception in save_group_or_entity, in was dublette"].append(f"Exception in save_group_or... :  {e}")
                # return ent

        else:
            log(f"Entity was no Person: {ent}, type:({type(ent)})")
            return ent

    def get_vorfin_or_entity(self, p):
        """
        Reverse helper functin of save_group_or_entity: returns the entity, or, 
        if it is a group, the newly created VORFIN.
        """
        log(f"in get_vorfin or entity, p is {p} ({type(p)})")
        if isinstance(p, Group):
            log(f"was instance of Group: {p}")
            try:
                group = Group.objects.get(id=p.id)
                log(f"returning group.vorfin: {group.vorfin} ({group.vorfin.id})")
                return group.vorfin
            except Exception as e:
                TempRel.errors["Could not find group error"].append(
                    f"Could not find group for: {p}({p.id})")
                logger.exception(e)
                res = None
                log(f"result set to {res}")
                return res
        else:
            log(f"in else. p is {p}, type is: {type(p)}")
            if not isinstance(p, AbstractEntity):
                msg = f"Expected entity to be subclass of AbstracEntity. Got {type(p)} instead."
                TempRel.errors["Entity Expectation Error"].append(msg)
                # raise Exception(msg)
            log(f"no exception, returning {p}.")
            return p

    @classmethod
    def re_create_rels(cls):
        """
        Main method for re-creating stored relations. Runs in bulk-processing over all instances of TempRel.
        """
        for inst in cls._instances:
            log(f"in re_create_rels. instance is {inst}. a is : {inst.a}, atpye {type(inst.a)}, b is {inst.b}, btype {type(inst.b)}")
            a = inst.get_vorfin_or_entity(inst.a)
            b = inst.get_vorfin_or_entity(inst.b)
            if a and b:
                if not inst.model.objects.filter(**inst.get_dict_data()).exists():
                    # TODO: inspect, CHANGED THIS
                    rel, c = inst.model.objects.get_or_create(
                        **inst.get_dict_data())
                    write_to_changed_flag = False
                    for x in [a, b]:
                        if cls.is_vorfin(x) and cls.vorfin_not_red_check(x):
                            write_to_changed_flag = True
                    if c:
                        cls._created_counter += 1
                        cls.created_by_model[inst.model]
                        cls.created[inst.model].append(rel)
                        log(
                            f"Re-created deleted relation: {rel}. type = {inst.model.__name__}")
                        if inst.ampel:
                            AmpelTemp.objects.get_or_create(
                                temp_entity=rel, status=inst.ampel)
                        log(f"Set ampel for new relation to: {inst.ampel}.")
                        if write_to_changed_flag:
                            cls.changed_edited_vorfins.append(rel)
                else:
                    log(f"Relation existed: {inst}")
                    cls.existed_by_model[inst.model]
                    cls.existed.append(inst)
            else:
                log(f"a or b was none: a:{a} b: {b}")

        log(f"Re-created {cls._created_counter} - Relations successfully.")

    @classmethod
    def log_stats_report(cls):
        """
        Helper to log stored statistics on created relations and those that already existed.
        """

        log(f"Number of relations that already existed = {len(cls.existed)}")
        log(f"Number of relations that wher re-created = {len(cls.created)}")
        log(f"created counter: {cls._created_counter}")

        log(f"created relations by model:")
        for mod, count in cls.created_by_model.items():
            log(f"\t{mod}: {count}")

        log(f"existed relations by model:")
        for mod, count in cls.existed_by_model.items():
            log(f"\t{mod}: {count}")

        for model, rels in cls.created.items():
            if model == PersonPerson:
                for r in rels:
                    kind = PPKind.classify(r)
                    log(f"Re-created PP-relation {r}\nkind:{kind}.\n")
            else:
                for r in rels:
                    log(
                        f"Re-created relation of type: {r.__class__.__name__}. rel is: {r}")

        log(f"Changed edited vorfins {len(cls.changed_edited_vorfins)}")

    @classmethod
    def get_changed_edited_vorfins_dataframe(cls):

        def get_vorfin_ampel(per):
            """
            Helper to get ampel of vorfin entry. If entity not a vorfin:
            returns 'NO VORFIN'. Only used to write the ampel status to 
            the excel file here.
            """
            if per in PersonHelper._vorfins:
                ampel = get_ampel(per)
                return ampel
            else:
                return "NO VORFIN"

        if cls.changed_edited_vorfins:
            data = []
            for rel in cls.changed_edited_vorfins:
                model = rel.__class__
                a_name = model.get_related_entity_field_nameA()
                b_name = model.get_related_entity_field_nameB()

                ent_a = getattr(rel, a_name)
                ent_a_ampel = get_vorfin_ampel(ent_a)
                ent_b = getattr(rel, b_name)
                ent_b_ampel = get_vorfin_ampel(ent_b)
                a_id = ent_a.id
                b_id = ent_b.id
                rt = rel.relation_type.name
                sd = rel.start_date_written
                ed = rel.end_date_written

                temp = {
                    "Relation": f"{model.__name__} ({rel.id})",
                    "source": f"{ent_a} ({ent_a.id})",
                    "source ampel": ent_a_ampel,
                    "relation_type": rt,
                    "target": f"{ent_b} ({ent_b.id})",
                    "target ampel": ent_b_ampel,
                    "StartDate": sd,
                    "EndDate": ed
                }
                data.append(temp)

            df = pd.DataFrame(data)
            return df

    @classmethod
    def get_created_rels_dataframe(cls):
        data = []
        for model, rels in cls.created.items():
            a_field = model.get_related_entity_field_nameA()
            b_field = model.get_related_entity_field_nameB()

            for r in rels:
                ent_a = getattr(r, a_field)
                ent_b = getattr(r, b_field)
                print(ent_a, type(ent_a))

                temp = {
                    "Relation": f"{model.__name__} ({r.id})",
                    "source": f"{ent_a} ({ent_a.id})",
                    "relation_type": r.relation_type.name,
                    "target": f"{ent_b} ({ent_b.id})",
                    "start": r.start_date_written,
                    "end": r.end_date_written,
                    "rel_class": "None"
                }
                # temp[a_field] = ent_a
                # temp[a_field+"_id"] = ent_a.id
                # temp[b_field] = ent_b
                # temp[b_field+"_id"] = ent_b.id

                # temp["source"] = ent_a
                # temp["source"+"_id"] = ent_a.id
                # temp["target"] = ent_b
                # temp["target"+"_id"] = ent_b.id

                if model == PersonPerson:
                    kind = PPKind.classify(r)
                    temp["rel_class"] = kind

                data.append(temp)

        df = pd.DataFrame(data)

        return df

    @classmethod
    def write_changed_edited_vorfins_to_excel(cls):
        """
        Writes collected relations in which already edited vorfins participate to an xlsx file.
        """
        df = cls.get_changed_edited_vorfins_dataframe()
        df.to_excel(
            "../../data/output/relations_that_changed_edited_vorfins.xlsx")

    @classmethod
    def write_created_to_excel(cls, filename="manually_created_rels_check.xlsx"):
        """
        Helper function to export the list of newly created relations as a xlsx - file. 
        """

        if not filename.endswith(".xlsx"):
            filename = filename + ".xlsx"

        df = cls.get_created_rels_dataframe()
        df.to_excel(f"../../data/output/{filename}", index=False)

    def __str__(self):
        return f"<TempRel: {self.a} - {self.rt} - {self.b}>"

    def __repr__(self):
        return self.__str__()



from django.test import TestCase
from dubletten_tool.src.classes import TempRel


class TempRelTest(TestCase):

    def setup(self, *args, **kwargs):
        print("setup test")

    def test_temp_rel_setup_and_reset(self, *args, **kwargs):

        # assert that the newly imported class is not setup yet, i.e. has no _instances attr
        self.assertRaises(AttributeError, getattr, TempRel, "_instances")
        TempRel.setup()
        self.assertEquals(TempRel._instances, [])



import logging 
import datetime
import os
import pathlib


"""
this file contains loggers to be used within the notebooks to create persistent documentation of complicated imports / changes. 


streamLogger: logs to the standard output. 
fileLogger: logs to a file given to the logger. The persistent option.
"""
streamLogger = logging.getLogger(name="NotebookStreamLogger")
fileLogger = logging.getLogger(name="NotebookFileLogger")

SH = logging.StreamHandler()
FH = logging.FileHandler(filename="../logs/test/output_test.txt")
fileLogger.addHandler(FH)
fileLogger.setLevel(logging.DEBUG)



def logger_factory(file_name, test=True, stream=True):

    timestamp = datetime.datetime.now().strftime("%Y-%m-%d (%H:%M)")
    logger = logging.getLogger(name="defaultStreamLogger")

    if test:
        path =  "../logs/test/"
    else:
        path =  "../logs/keep/"

    filepath = path + file_name+"_"+timestamp+".txt"
    fileHandler = logging.FileHandler(filepath)
    if stream:
        streamHandler = logging.StreamHandler()
        logger.addHandler(streamHandler)
    
    logger.addHandler(fileHandler)
    logger.setLevel(logging.DEBUG)

    return logger
    


from django.core.management.base import BaseCommand, CommandError
import pandas as pd
import itertools as it
from collections import defaultdict, Counter
import json
from dubletten_tool.models import PersonProxy, Group, Suggestions, StatusButtonProxyType, StatusButtonGroupType, StatusButtonGroup, StatusButtonProxyType, StatusButtonProxy
from apis_core.apis_entities.models import Person
from apis_core.apis_metainfo.models import Collection
import logging
from django.contrib.auth.models import User
from dubletten_tool.logger import init_logger


log = logging.getLogger("DBLogger")

user = User.objects.get(username="GPirgie")

class Command(BaseCommand):

    def run_import_groups(self):
        #delete existing PersonProxy, Group and Button Objects
        PersonProxy.objects.all().delete()
        Group.objects.all().delete()
        StatusButtonGroupType.objects.all().delete()
        StatusButtonProxyType.objects.all().delete()
        StatusButtonGroup.objects.all().delete()
        StatusButtonProxy.objects.all().delete()
        
        ACC = Collection.objects.get(name="Import ACCESS full 13-10-21")
        SELECTION = Person.objects.exclude(collection__in=[ACC])
        
        df_per = pd.DataFrame(SELECTION.values("id","name", "first_name", "gender", "start_date", "end_date", "start_date_written", "end_date_written")).set_index("id")
        per = df_per[["name", "first_name", "gender", "start_date", "end_date", "start_date_written", "end_date_written"]]
        #per["fullname"] = per.name + ", " + per.first_name
        per["sd"] = [str(d) for d in per["start_date"]]
        per["ed"] = [str(d) for d in per["end_date"]]

        self.stdout.write("Calculating Groups Started")
        for idx, (d, group) in enumerate(per.groupby(["name", "first_name", "gender", "sd", "ed"])):
                    bd, dd = d[3], d[4]
                    bd = bd.replace("None", "")
                    dd = dd.replace("None", "")
                    if len(group) > 1:
                        if bd and dd:
                            name = (f"{d[0]}, {d[1]} [B+D]")
                        elif bd:
                            name = (f"{d[0]}, {d[1]} [B]")
                        elif dd:
                            name = (f"{d[0]}, {d[1]} [D]")
                        else:
                            name = (f"{d[0]}, {d[1]}")

                        g, c = Group.objects.get_or_create(name=name)
                        log.info(f"CREATED GROUP {g.id}, with name {g.name}", extra={"user":user, "action": "SETUP GROUP: Created Group"})
                        index = group.index.to_list()
                        for ix in index:
                            gm, c = PersonProxy.objects.get_or_create(person=Person.objects.get(id=ix))
                            gm.status="candidate"
                            gm.save()
                            g.members.add(gm)
                        g._gender = d[2]
                        g.save()
                        log.info(f"GROUP {g.id}, with name {g.name} has members {[{'proxy_id':p.id, 'person_id':p.person.id, 'status':p.status} for p in g.members.all()]}", extra={"user":user, "action": "SETUP GROUP: Added Members"})


                    elif len(group) == 1:
                        s = group.index.to_list()[0]
                        per = Person.objects.get(id=s)
                        prox, c = PersonProxy.objects.get_or_create(person=per)
                        prox.status = "single"
                        prox.save()
                        log.info(f"CREATED SINGLE {prox.id} with name {prox.name} and person_id {prox.person.id} with status {prox.status}", extra={"user":user, "action": "SETUP GROUP: Created Single"})
        self.stdout.write("Finished: Calculating Groups")


    def create_ampel_buttons(self):
        self.stdout.write(f"Starting Create Ampel Buttons")
        types = [("Checked Members", "CM" ),("Checked Suggestions", "CSu" ),("Checked Singles", "CSi"),("Ready To Merge", "RM")]
        for el in types:
            btn, c = StatusButtonGroupType.objects.get_or_create(name=el[0], short=el[1])
            btn.add_to_all_groups()
            self.stdout.write(f"Added Button '{el[0]} to all Groups.")
        


    def calculate_names(self):
        self.stdout.write("Started: Calculate Names")
        
        for p in PersonProxy.objects.all():
            p._names = p.names_list
            p._first_names = p.first_names_list
            p.save()
        
        sn = {tuple(s.allnames) for s in PersonProxy.objects.all()}
        sn = [set(el) for el in sn]
        all_sn = {part for el in sn for part in el}
        
        sf = {tuple(s.allfirst_names) for s in PersonProxy.objects.all()}
        sf = [set(el) for el in sf]
        all_sf = {part for el in sf for part in el}


        p_n = defaultdict(list)
        p_f = defaultdict(list)
        [p_n[tuple(p.allnames)].append(p.id) for p in PersonProxy.objects.all()]    
        [p_f[tuple(p.allfirst_names)].append(p.id) for p in PersonProxy.objects.all()]    
        
        self.stdout.write("step 1 finished")
        
        def get_names(names_set, res_dic):
            self.stdout.write("running get names")
            res = defaultdict(list)
            for n in names_set:
                for k in res_dic:
                    if n in k:
                        res[n] += res_dic[k]
            return res
        
        res_n = get_names(all_sn, p_n)
        res_f = get_names(all_sf, p_f)
        
        self.stdout.write("step 2 finished, get names done")
        
        def get_per_dic(data):
            self.stdout.write("running get_per_dic")
            per_dic = defaultdict(set)
            for idx, i in enumerate(PersonProxy.objects.values_list("id")):
                test = i[0]
                for k, v in data.items():
                    if test in v:
                        per_dic[test] = per_dic[test].union(set(v))
                        per_dic[test].remove(test)
            return per_dic
        
        n = get_per_dic(res_n)
        f = get_per_dic(res_f)
        self.stdout.write("finished all calculations")
        
        full_res = defaultdict(set)

        for k,v in n.items():
        
            v2 = f[k]
            res = v.intersection(v2)
            if res:
                full_res[k] = res
            else:
                full_res[k] = set()
        
        data = {k:list(v) for k, v in full_res.items()}
        # with open("/Users/gregorpirgie/Desktop/full_res_new.json", "w") as file:
        #     data = json.dumps({k:list(v) for k, v in full_res.items()})
        #     file.write(data)
        self.stdout.write("Finished: Calculate Names")

        return data
        

  
    def handle(self, *args, **options):
        self.run_import_groups()
        self.create_ampel_buttons()
        data = self.calculate_names()
        #data = json.loads(data)
        Suggestions.objects.all().delete()
        Suggestions.objects.get_or_create(data=data)
        self.stdout.write("Finished: ALL PROCESSES")





from . import PPHelper, PersonType, Enum, Group


class PPKind(Enum):
    # expect relation     VORVOR = "vorfinA - vorfinB"
    DUBDUB = "dubletteA - dubletteB"
    DUBSAME = "dubletteA - dubletteA"  # expect relation    VORSAME
    DUBSING = "dublette - single"  # expect relation VORSING
    SINGDUB = "single - dublette"  # expect relatio SINGVOR
    VORVOR = "vorfinA - vorfinB"  # outcome of DUBDUB
    VORSAME = "vorfinA - vorfinA"  # outcome of DUBSAME
    SINGVOR = "single - vorfin"  # outcome of SINGDUB
    VORSING = "vorfin - single"  # outcome of DUBSING
    SINGSING = "single - single"  # should not occur in script
    # should not occur in script, or how could this happen? should be VORVOR or VORSAME
    DUBVOR = "dublette - vorfin"
    # should not occur in script; should be VORVOR or VORSAME
    VORDUB = "vorfin - dublette"
    SINSING = "singleA - singleB"  # should not occur in script
    SINGSAME = "singleA - singleA"  # should not occur in script

    @classmethod
    def classify(cls, rel, in_step="external call", strict=False):
        """
        Main Method to classify a given PersonPerson relation, 
        by classifying the participating entities (persons) first.

        @param strict: if set to True, throws an error if a classification does not conform to our expectations. 
        --> must be False when processing manually created relations; we know, that they contain 
        classifications that we want to get rid of, but we want them to be logged and written first.

        """

        perA = rel.related_personA
        perB = rel.related_personB
        atype = PersonType.classify(perA, strict=strict)
        btype = PersonType.classify(perB, strict=strict)

        if atype == PersonType.DUBLETTE == btype:
            aprox = perA.personproxy
            bprox = perB.personproxy
            agroup = Group.objects.get(members__in=[aprox])
            bgroup = Group.objects.get(members__in=[bprox])

            if agroup == bgroup:
                return cls.DUBSAME
            else:
                return cls.DUBDUB

        # This part is new to catch also VORSAME in classification
        elif atype == PersonType.VORFIN == btype:
            if perA == perB:
                return cls.VORSAME
            else:
                return cls.VORVOR
        # end of new part.

        # last addition:
        elif atype == PersonType.SINGLE == btype:
            if perA == perB:
                return cls.SINGSAME
            else:
                return cls.SINGSING

        # end of last addition

        else:
            res = PPHelper._kindmap.get((atype, btype), None)
            if res:
                if res in [cls.SINGSING, cls.DUBVOR, cls.VORDUB]:
                    msg = f"In rel classify '{in_step}': res was {res}. This should not happen. Relation was: {rel} ({rel.id})."
                    PPHelper.enum_errors["Strict classification error"].append(
                        msg)
                    if strict:
                        raise Exception(msg)
                    # note: changed this here I. My guess is: the error occurs after the relation has been created.
                    # Therefore: this didn't return before and so there was no kind. This prevented the
                    # expectation missmatch to kick in .

                return res

            else:
                msg = f"Could not classify relation: {rel}. Res was: {res}.\nTypes where a:{atype}, b:{btype}."
                PPHelper.enum_errors["Classification error: No result"].append(
                    msg)
                raise Exception(msg)





from django.core.management.base import BaseCommand
from dubletten_tool.models import  Group
from apis_core.apis_metainfo.models import Collection
import logging
from django.contrib.auth.models import User
from dubletten_tool.logger import init_logger
from apis_core.apis_metainfo.models import Collection
from dubletten_tool.models import  Group
from dubletten_tool.merge_functions import write_person_person_rels, MergeGroup

log = logging.getLogger("DBLogger")

user = User.objects.get(username="GPirgie")


class Command(BaseCommand):

    def handle(self, *args, **kwargs):
        merged, c = Collection.objects.get_or_create(name="Vorfinale Einträge")
        merged.tempentityclass_set.all().delete()

        gg = Group.objects.all()
        all = len(gg)
        for idx, g in enumerate(gg):
            MergeGroup(g).run_process()
            print(f"Create Group: {idx}/{all}")
        
        for idx, g in enumerate(gg):
            write_person_person_rels(g)            
            print(f"Write PerPerRelation: {idx}/{all}")

        print("all processes finished with success")





from . import Enum, PersonHelper

class PersonType(Enum):
    """
    Enum that has a single method 'classify' in which it classifies each 
    Person into the three types. Used in PPKind to classify each relation by 
    the kinds of persons participating in it.
    """
    SINGLE = "single"
    DUBLETTE = "dublette"
    VORFIN = "vorfin"


    @classmethod
    def classify(cls, p, strict=True):
        """
        Main classification of a given person.
        
        @param strict: if True, throws an Error if Person is not in singles, dubletten or vorfins. 
        
        --> use strict mode when proceessing relations coming from dubletten (i.e. in re-merge).
        --> disable strict when processing all relations in a vorfin, to catch and process
        the manually created relations that might contain participating Persons outside 
        of our initial de-duplication scope. those are simply classified as singles, because we 
        haven't processed them yet. 
        """
        if p in PersonHelper._singles:
            return PersonType.SINGLE
        elif p in PersonHelper._dubletten:
            return PersonType.DUBLETTE
        elif p in PersonHelper._vorfins:
            return PersonType.VORFIN
        else:
            if strict:
                msg = f"couldn't find type for person {p}({p.id}) {p.__class__.__name__}. Was neither in singles, dubletten nor vorfins"
                PersonHelper.enum_errors["PersonType Classification Error"].append(msg)
                raise Exception(msg)
            else:
                return PersonType.SINGLE




from django.core.management.base import BaseCommand, CommandError
import pandas as pd
import itertools as it
from collections import defaultdict, Counter
import json
from dubletten_tool.models import PersonProxy, Group, Suggestions, StatusButtonProxyType, StatusButtonGroupType, StatusButtonGroup, StatusButtonProxyType, StatusButtonProxy
from apis_core.apis_entities.models import Person
from apis_core.apis_metainfo.models import Collection
import logging
from django.contrib.auth.models import User
from dubletten_tool.logger import init_logger


log = logging.getLogger("DBLogger")

user = User.objects.get(username="GPirgie")

class Command(BaseCommand):

    def create_new_groups(self, per_set):
        selection = Person.objects.filter(id__in=[p.id for p in per_set])
        df_per = pd.DataFrame(selection.values("id","name", "first_name", "gender", "start_date", "end_date", "start_date_written", "end_date_written")).set_index("id")
        per = df_per[["name", "first_name", "gender", "start_date", "end_date", "start_date_written", "end_date_written"]]
        per["sd"] = [str(d) for d in per["start_date"]]
        per["ed"] = [str(d) for d in per["end_date"]]

        self.stdout.write("Calculating Groups Started")
        for idx, (d, group) in enumerate(per.groupby(["name", "first_name", "gender", "sd", "ed"])):
            bd, dd = d[3], d[4]
            bd = bd.replace("None", "")
            dd = dd.replace("None", "")
            if len(group) > 1:
                if bd and dd:
                    name = (f"{d[0]}, {d[1]} [B+D]")
                elif bd:
                    name = (f"{d[0]}, {d[1]} [B]")
                elif dd:
                    name = (f"{d[0]}, {d[1]} [D]")
                else:
                    name = (f"{d[0]}, {d[1]}")
                    
                name += " {ACC}"

                g, c = Group.objects.get_or_create(name=name)
                log.info(f"CREATED GROUP {g.id}, with name {g.name}", extra={"user":user, "action": "SETUP GROUP: Created Group"})
                index = group.index.to_list()
                for ix in index:
                    gm, c = PersonProxy.objects.get_or_create(person=Person.objects.get(id=ix))
                    gm.status="candidate"
                    gm.save()
                    g.members.add(gm)
                g._gender = d[2]
                g.save()
                log.info(f"GROUP {g.id}, with name {g.name} has members {[{'proxy_id':p.id, 'person_id':p.person.id, 'status':p.status} for p in g.members.all()]}", extra={"user":user, "action": "SETUP GROUP: Added Members"})


            elif len(group) == 1:
                s = group.index.to_list()[0]
                per = Person.objects.get(id=s)
                prox, c = PersonProxy.objects.get_or_create(person=per)
                prox.status = "single"
                prox.save()
                log.info(f"CREATED SINGLE {prox.id} with name {prox.name} and person_id {prox.person.id} with status {prox.status}", extra={"user":user, "action": "SETUP GROUP: Created Single"})
        self.stdout.write("Finished: Calculating Groups")

    # copied over from management commands
    def create_ampel_buttons(self):
        self.stdout.write(f"Starting Create Ampel Buttons")
        types = [("Checked Members", "CM" ),("Checked Suggestions", "CSu" ),("Checked Singles", "CSi"),("Ready To Merge", "RM")]
        for el in types:
            btn, c = StatusButtonGroupType.objects.get_or_create(name=el[0], short=el[1])
            btn.add_to_all_groups()
            self.stdout.write(f"Added Button '{el[0]} to all Groups.")


    # copied over from management commands
    def calculate_names(self):
        #self.stdout.write("Started: Calculate Names")

        for p in PersonProxy.objects.all():
            p._names = p.names_list
            p._first_names = p.first_names_list
            p.save()

        sn = {tuple(s.allnames) for s in PersonProxy.objects.all()}
        sn = [set(el) for el in sn]
        all_sn = {part for el in sn for part in el}

        sf = {tuple(s.allfirst_names) for s in PersonProxy.objects.all()}
        sf = [set(el) for el in sf]
        all_sf = {part for el in sf for part in el}


        p_n = defaultdict(list)
        p_f = defaultdict(list)
        [p_n[tuple(p.allnames)].append(p.id) for p in PersonProxy.objects.all()]    
        [p_f[tuple(p.allfirst_names)].append(p.id) for p in PersonProxy.objects.all()]    

        #self.stdout.write("step 1 finished")

        def get_names(names_set, res_dic):
            #self.stdout.write("running get names")
            res = defaultdict(list)
            for n in names_set:
                for k in res_dic:
                    if n in k:
                        res[n] += res_dic[k]
            return res

        res_n = get_names(all_sn, p_n)
        res_f = get_names(all_sf, p_f)

        #self.stdout.write("step 2 finished, get names done")

        def get_per_dic(data):
            #self.stdout.write("running get_per_dic")
            per_dic = defaultdict(set)
            for idx, i in enumerate(PersonProxy.objects.values_list("id")):
                test = i[0]
                for k, v in data.items():
                    if test in v:
                        per_dic[test] = per_dic[test].union(set(v))
                        per_dic[test].remove(test)
            return per_dic

        n = get_per_dic(res_n)
        f = get_per_dic(res_f)
        #self.stdout.write("finished all calculations")

        full_res = defaultdict(set)

        for k,v in n.items():

            v2 = f[k]
            res = v.intersection(v2)
            if res:
                full_res[k] = res
            else:
                full_res[k] = set()

        data = {k:list(v) for k, v in full_res.items()}
        # with open("/Users/gregorpirgie/Desktop/full_res_new.json", "w") as file:
        #     data = json.dumps({k:list(v) for k, v in full_res.items()})
        #     file.write(data)
        #self.stdout.write("Finished: Calculate Names")

        return data


  
    def handle(self, *args, **options):
        nopps = [p for p in Person.objects.all() if not hasattr(p, "personproxy")]
        no_vorfins_no_proxies = [p for p in nopps if not "vorfin" in p.name]
        self.create_new_groups(no_vorfins_no_proxies)
        self.create_ampel_buttons()
        data = self.calculate_names()
        Suggestions.objects.all().delete()
        Suggestions.objects.get_or_create(data=data)
        self.stdout.write("All processes finished.")


    





# Generated by Django 3.2 on 2022-10-26 21:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dubletten_tool', '0004_auto_20221026_1855'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ampel',
            name='status',
            field=models.CharField(choices=[('red', 'red'), ('yellow', 'yellow'), ('green', 'green')], default='red', max_length=300),
        ),
    ]



from django.shortcuts import render
from django.http import HttpResponse, JsonResponse
from django.views.generic import View, TemplateView
from django.utils.decorators import method_decorator
# Create your views here.
from django.template.loader import render_to_string
from django.contrib.auth.decorators import login_required, user_passes_test
#from .grouping_script_1 import run
from .models import Group, PersonProxy, StatusButtonGroup, StatusButtonProxy, StatusButtonGroupType
import json
from .suggestions import get_levensthein, check_all_names, get_name_suggestions
from .forms import GroupForm, PersonProxyForm
import logging
from .logger import init_logger
from .merge_tables import VorfinListTable
from apis_core.apis_metainfo.models import Collection
import re
# copied over, delete unnecessary parts

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.http import HttpResponse
from django.shortcuts import redirect, get_object_or_404
from django.template.loader import select_template
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views import View
from django.views.generic import DeleteView
from django_tables2 import RequestConfig
from guardian.core import ObjectPermissionChecker
from reversion.models import Version
import importlib

from apis_core.apis_entities.forms import (
    GenericFilterFormHelper,
    NetworkVizFilterForm,
    PersonResolveUriForm,
    GenericEntitiesStanbolForm,
)
from apis_core.apis_entities.models import Place, Person, Institution
from apis_core.apis_entities.tables import get_entities_table

from apis_core.apis_entities.models import AbstractEntity
from apis_core.apis_labels.models import Label
from apis_core.apis_metainfo.models import Uri
from apis_core.apis_relations.models import AbstractRelation, PersonInstitution, PersonPerson
from apis_core.apis_relations.tables import LabelTableEdit
from apis_core.apis_entities.forms import get_entities_form, FullTextForm, GenericEntitiesStanbolForm
from apis_core.apis_entities.views import get_highlighted_texts
from apis_core.apis_entities.views import set_session_variables
from apis_core.apis_vocabularies.models import TextType

from django.conf import settings
from django.contrib.auth.mixins import UserPassesTestMixin
from django.db.models import Q
from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.template.loader import select_template
from django.views import View
from django_tables2 import RequestConfig
from django.urls import reverse


from apis_core.apis_vocabularies.models import PersonInstitutionRelation, PersonPersonRelation
from apis_core.apis_relations.models import AbstractRelation
# , EntityDetailViewLabelTable,  get_generic_relations_table,
from apis_core.apis_relations.tables import LabelTableBase
from apis_core.helper_functions.utils import access_for_all
from apis_core.apis_entities.models import AbstractEntity
from apis_core.apis_entities.views import get_highlighted_texts
import json

from django.conf import settings
from django.contrib.auth.decorators import login_required, user_passes_test
from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.contenttypes.models import ContentType
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.shortcuts import render, redirect, get_object_or_404
from django.urls import reverse
from django.urls import reverse_lazy
from django_tables2 import RequestConfig
from django_tables2 import SingleTableView
from django_tables2.export.views import ExportMixin
#from .models import Ampel

if "charts" in settings.INSTALLED_APPS:
    from charts.models import ChartConfig
    from charts.views import create_payload
# from reversion_compare.views import HistoryCompareDetailView

from apis_core.apis_metainfo.models import Uri, UriCandidate, Text
from apis_core.apis_relations.models import AbstractRelation
from apis_core.helper_functions.RDFParser import RDFParser
from apis_core.helper_functions.stanbolQueries import retrieve_obj
from apis_core.helper_functions.utils import (
    access_for_all,
    access_for_all_function,
    ENTITIES_DEFAULT_COLS,
)
from apis_core.apis_entities.filters import get_list_filter_of_entity
from .merge_functions import write_person_person_rels, MergeGroup, update_merging_collections
#from .merge_filters import get_list_filter_of_entity
#
#  DEactivated, check if this works
from .merge_tables import get_generic_relations_table

from dubletten_tool.src.functions.remerge_single_group import remerge_single_group


if "apis_ampel" in settings.INSTALLED_APPS:
    from apis_ampel.helper_functions import is_ampel_active


@method_decorator(login_required, name="dispatch")
class UpdateCollectionsView(View):
    def get(self, request, *args, **kwargs):
        msg = "Update Collections Finished."
        success = True

        try:
            update_merging_collections()
        except Exception as e:
            msg = f"Something went wrong:\n{e}"
            success = False

        return JsonResponse({"msg": msg, "success": success})


@method_decorator(login_required, name="dispatch")
class RemergeGroupViewOLD(View):

    def get(self, request, *args, **kwargs):

        g_id = kwargs.get("g_id")
        try:
            group = Group.objects.get(id=g_id)
            vorfin = group.vorfin
            vorfin.delete()
        except Exception as e:
            print(e)

        try:
            MergeGroup(group).run_process()
            write_person_person_rels(group)
            msg = f"Success for {g_id}.\nNew vorfinaler Eintrag is: {group.vorfin} [{group.vorfin.id}]"

        except Exception as e:
            msg = f"merge failed for group_id: {g_id}, error was:\n {e}"

        return JsonResponse({"msg": msg})


@method_decorator(login_required, name="dispatch")
class RemergeGroupView(View):

    # TODO: extract get display return logic into seperate function
    # TODO: perserve old group name on remerge.
    # TODO: maybe save old vorfin data / rendered detail view and pass to frontend (display on button click)
    # TODO: but would need to disable all buttons, links, etc. Cumbersome.

    @staticmethod
    def get_remerge_data(group, vorfin_list=None, *args, **kwargs):

        # Single group logic
        vorfin = group.vorfin
        if vorfin:
            old_vorfin_name = group.vorfin.name
            old_vorfin_id = group.vorfin.id
        else:
            old_vorfin_name = None
            old_vorfin_id = None

        try:
            res_dict = remerge_single_group(group, vorfins=vorfin_list)
            new_vorfin = res_dict.get("new_vorfin")
            print(res_dict)
            msg = f"Merged {group} ({group.name}) into new vorfin: {new_vorfin} ({new_vorfin.id}).\nDeleted old vorfin: {old_vorfin_name} ({old_vorfin_id})."
        except Exception as e:
            msg = f"merge failed for group_id: {group.id}, error was:\n {e}"
            raise Exception(msg)

            # return HttpResponse(f"{ELM.log_report()}\n\n{ELM.log_details()}")

        def df_to_html(df):
            if df is None or df.empty:
                return "<p class='text-center'>Nothing to Display.</p>"
            else:
                return df.to_html(classes="table table-sm", border=None, justify="left", index=False)

        result = {
            "msg": msg,
            "rels_changed": {
                "title": "Relations with changed vorfins",
                "table": df_to_html(res_dict.get("changed_vorfins")),
            },
            "rels_added": {
                "title": "Re-added Relations",
                "table": df_to_html(res_dict.get("created_rels")),
            },
            "new_vorfin": str(res_dict["new_vorfin"]),
            "new_vorfin_id": res_dict["new_vorfin_id"],
        }

        return result

    def get(self, request, *args, **kwargs):
        print("called new remerge.")
        g_id = kwargs.get("g_id")
        try:
            group = Group.objects.get(id=g_id)

            result = RemergeGroupView.get_remerge_data(group)
        except Exception as e:
            print(e)
            return JsonResponse({"msg": f"Could not find Group object with id {g_id}. ReMerge Aborted."})

        # return HttpResponse(test)
        return JsonResponse(result)


@method_decorator(login_required, name="dispatch")
class GenericEntitiesDetailView(UserPassesTestMixin, View):

    def test_func(self):
        access = access_for_all(self, viewtype="detail")
        return access

    def get(self, request, *args, **kwargs):

        entity = kwargs['entity'].lower()
        pk = kwargs['pk']
        entity_model = AbstractEntity.get_entity_class_of_name(entity)
        instance = get_object_or_404(entity_model, pk=pk)
        merged_rel = PersonPersonRelation.objects.get(name="data merged from")
        persons_ = [r.related_personB for r in PersonPerson.objects.filter(
            related_personA=instance, relation_type=merged_rel)]

        relations = AbstractRelation.get_relation_classes_of_entity_name(
            entity_name=entity)
        side_bar = []
        for rel in relations:
            match = [
                rel.get_related_entity_classA().__name__.lower(),
                rel.get_related_entity_classB().__name__.lower()
            ]
            prefix = "{}{}-".format(match[0].title()[:2], match[1].title()[:2])
            print("relations_table class is:", rel)
            table = get_generic_relations_table(
                relation_class=rel, entity_instance=instance, detail=True)
            if match[0] == match[1]:
                title_card = entity.title()
                dict_1 = {'related_' + entity.lower() + 'A': instance}
                dict_2 = {'related_' + entity.lower() + 'B': instance}
                if 'apis_highlighter' in settings.INSTALLED_APPS:
                    objects = rel.objects.filter_ann_proj(request=request).filter_for_user().filter(
                        Q(**dict_1) | Q(**dict_2))
                else:
                    objects = rel.objects.filter(
                        Q(**dict_1) | Q(**dict_2))
                    if callable(getattr(objects, 'filter_for_user', None)):
                        objects = objects.filter_for_user()
            else:
                if match[0].lower() == entity.lower():
                    title_card = match[1].title()
                else:
                    title_card = match[0].title()
                dict_1 = {'related_' + entity.lower(): instance}
                if 'apis_highlighter' in settings.INSTALLED_APPS:
                    objects = rel.objects.filter_ann_proj(
                        request=request).filter_for_user().filter(**dict_1)
                else:
                    objects = rel.objects.filter(**dict_1)
                    if callable(getattr(objects, 'filter_for_user', None)):
                        objects = objects.filter_for_user()
            tb_object = table(data=objects, prefix=prefix)
            tb_object_open = request.GET.get(prefix + 'page', None)
            RequestConfig(request, paginate={
                          "per_page": 10}).configure(tb_object)
            side_bar.append(
                (title_card, tb_object, ''.join(
                    [x.title() for x in match]), tb_object_open)
            )
        object_lod = Uri.objects.filter(entity=instance)
        object_texts, ann_proj_form = get_highlighted_texts(request, instance)
        object_labels = Label.objects.filter(temp_entity=instance)
        tb_label = LabelTableBase(
            data=object_labels, prefix=entity.title()[:2]+'L-')
        tb_label_open = request.GET.get('PL-page', None)
        side_bar.append(('Label', tb_label, 'PersonLabel', tb_label_open))
        RequestConfig(request, paginate={"per_page": 10}).configure(tb_label)
        template = select_template([
            'merge/detail_views/entity_detail_generic.html'
        ])
        tei = getattr(settings, "APIS_TEI_TEXTS", [])
        if tei:
            tei = set(tei) & set([x.kind.name for x in instance.text.all()])
        ceteicean_css = getattr(settings, "APIS_CETEICEAN_CSS", None)
        ceteicean_js = getattr(settings, "APIS_CETEICEAN_JS", None)
        openseadragon_js = getattr(settings, "APIS_OSD_JS", None)
        openseadragon_img = getattr(settings, "APIS_OSD_IMG_PREFIX", None)
        iiif_field = getattr(settings, "APIS_IIIF_WORK_KIND", None)
        if iiif_field:
            try:
                if "{}".format(instance.kind) == "{}".format(iiif_field):
                    iiif = True
                else:
                    iiif = False
            except AttributeError:
                iiif = False
        else:
            iiif = False
        iiif_server = getattr(settings, "APIS_IIIF_SERVER", None)
        iiif_info_json = instance.name
        try:
            no_merge_labels = [
                x for x in object_labels if not x.label_type.name.startswith('Legacy')
            ]
        except AttributeError:
            no_merge_labels = []

        nexturl = instance.get_next_url()
        if nexturl:
            nexturl = nexturl.replace("apis/entities", "dubletten"),
        prevurl = instance.get_prev_url()
        if prevurl:
            prevurl = prevurl.replace("apis/entities", "dubletten")

        return HttpResponse(template.render(
            request=request, context={
                'titles': instance.title.all(),
                'persons': persons_,
                'entity_type': entity,
                'object': instance,
                'right_card': side_bar,
                'no_merge_labels': no_merge_labels,
                'object_lables': object_labels,
                'object_texts': object_texts,
                'object_lod': object_lod,
                'tei': tei,
                'ceteicean_css': ceteicean_css,
                'ceteicean_js': ceteicean_js,
                'iiif': iiif,
                'openseadragon_js': openseadragon_js,
                'openseadragon_img': openseadragon_img,
                'iiif_field': iiif_field,
                'iiif_info_json': iiif_info_json,
                'iiif_server': iiif_server,
                'edit_url': instance.get_edit_url().replace("apis/entities", "dubletten"),
                'next_url': nexturl,
                'prev_url': prevurl,
            }
        ))


class GenericListViewNew(UserPassesTestMixin, ExportMixin, SingleTableView):
    formhelper_class = GenericFilterFormHelper
    context_filter_name = "filter"
    paginate_by = 25
    template_name = getattr(
        settings, "APIS_LIST_VIEW_TEMPLATE", "apis:apis_entities/generic_list.html"
    )
    template_name = "merge/generic_list.html"
    #login_url = "/accounts/login/"

    def get_model(self):
        model = ContentType.objects.get(
            app_label__startswith="apis_", model=self.entity.lower()
        ).model_class()
        return model

    def test_func(self):
        access = access_for_all(self, viewtype="list")
        if access:
            self.request = set_session_variables(self.request)
        return access

    def get_queryset(self, **kwargs):
        # note __g.pirgie__ changed the queryset for this view to only list persons from the deduplication collection
        # dedup collection called 'Vorfinale Einträge' in VieCPro
        self.entity = "Person"  # changed for merge
        vorfin = Collection.objects.get(name="Vorfinale Einträge")

        qs = (
            ContentType.objects.get(
                app_label__startswith="apis_", model=self.entity.lower()
            )
            .model_class()
            .objects.filter(collection__in=[vorfin])
        )

        self.filter = get_list_filter_of_entity(self.entity.title())(
            self.request.GET, queryset=qs
        )

        self.filter.form.helper = self.formhelper_class()
        return self.filter.qs

    def get_table(self, **kwargs):
        session = getattr(self.request, "session", False)
        entity = "Person"
        selected_cols = self.request.GET.getlist("columns")
        if session:
            edit_v = self.request.session.get("edit_views", False)

        else:
            edit_v = False
        if "table_fields" in settings.APIS_ENTITIES[entity.title()]:
            default_cols = settings.APIS_ENTITIES[entity.title(
            )]["table_fields"]
        else:
            default_cols = ["name"]
        default_cols = default_cols + selected_cols
        self.table_class = VorfinListTable

        # get_entities_table(
        #     self.entity.title(), edit_v, default_cols=default_cols
        # )
        # deleted VorfinListTable Here
        table = super(GenericListViewNew, self).get_table()
        RequestConfig(
            self.request, paginate={"page": 1, "per_page": self.paginate_by}
        ).configure(table)
        return table

    def get_context_data(self, **kwargs):
        model = self.get_model()
        context = super(GenericListViewNew, self).get_context_data()
        context[self.context_filter_name] = self.filter
        context["entity"] = self.entity
        context["app_name"] = "apis_entities"
        entity = self.entity.title()
        context["entity_create_stanbol"] = GenericEntitiesStanbolForm(
            self.entity)
        if "browsing" in settings.INSTALLED_APPS:
            from browsing.models import BrowsConf

            context["conf_items"] = list(
                BrowsConf.objects.filter(model_name=self.entity).values_list(
                    "field_path", "label"
                )
            )
        context["docstring"] = "{}".format(model.__doc__)
        if model._meta.verbose_name_plural:
            context["class_name"] = "{}".format(
                model._meta.verbose_name.title())
        else:
            if model.__name__.endswith("s"):
                context["class_name"] = "{}".format(model.__name__)
            else:
                context["class_name"] = "{}s".format(model.__name__)
        try:
            context["get_arche_dump"] = model.get_arche_dump()
        except AttributeError:
            context["get_arche_dump"] = None
        try:
            context["create_view_link"] = model.get_createview_url()
        except AttributeError:
            context["create_view_link"] = None
        if "charts" in settings.INSTALLED_APPS:
            app_label = model._meta.app_label
            filtered_objs = ChartConfig.objects.filter(
                model_name=model.__name__.lower(), app_name=app_label
            )
            context["vis_list"] = filtered_objs
            context["property_name"] = self.request.GET.get("property")
            context["charttype"] = self.request.GET.get("charttype")
            if context["charttype"] and context["property_name"]:
                qs = self.get_queryset()
                chartdata = create_payload(
                    context["entity"],
                    context["property_name"],
                    context["charttype"],
                    qs,
                    app_label=app_label,
                )
                context = dict(context, **chartdata)
        try:
            context["enable_merge"] = settings.APIS_ENTITIES[entity.title()
                                                             ]["merge"]
        except KeyError:
            context["enable_merge"] = False
        try:
            togg_cols = settings.APIS_ENTITIES[entity.title(
            )]["additional_cols"]
        except KeyError:
            togg_cols = []
        if context["enable_merge"] and self.request.user.is_authenticated:
            togg_cols = togg_cols + ["merge"]
        context["togglable_colums"] = togg_cols + ENTITIES_DEFAULT_COLS

        return context

    def render_to_response(self, context, **kwargs):
        download = self.request.GET.get("sep", None)
        if download and "browsing" in settings.INSTALLED_APPS:
            import datetime
            import time
            import pandas as pd

            sep = self.request.GET.get("sep", ",")
            timestamp = datetime.datetime.fromtimestamp(time.time()).strftime(
                "%Y-%m-%d-%H-%M-%S"
            )
            filename = "export_{}".format(timestamp)
            response = HttpResponse(content_type="text/csv")
            if context["conf_items"]:
                conf_items = context["conf_items"]
                try:
                    df = pd.DataFrame(
                        list(
                            self.get_queryset().values_list(
                                *[x[0] for x in conf_items])
                        ),
                        columns=[x[1] for x in conf_items],
                    )
                except AssertionError:
                    response[
                        "Content-Disposition"
                    ] = 'attachment; filename="{}.csv"'.format(filename)
                    return response
            else:
                response[
                    "Content-Disposition"
                ] = 'attachment; filename="{}.csv"'.format(filename)
                return response
            if sep == "comma":
                df.to_csv(response, sep=",", index=False)
            elif sep == "semicolon":
                df.to_csv(response, sep=";", index=False)
            elif sep == "tab":
                df.to_csv(response, sep="\t", index=False)
            else:
                df.to_csv(response, sep=",", index=False)
            response["Content-Disposition"] = 'attachment; filename="{}.csv"'.format(
                filename
            )
            return response
        else:
            response = super(GenericListViewNew,
                             self).render_to_response(context)
            return response


@method_decorator(login_required, name='dispatch')
class GenericEntitiesEditView(View):

    def get(self, request, *args, **kwargs):
        entity = kwargs['entity']
        pk = kwargs['pk']
        entity_model = AbstractEntity.get_entity_class_of_name(entity)
        instance = get_object_or_404(entity_model, pk=pk)
        merged_rel = PersonPersonRelation.objects.get(name="data merged from")
        persons_ = [r.related_personB for r in PersonPerson.objects.filter(
            related_personA=instance, relation_type=merged_rel)]
        request = set_session_variables(request)
        relations = AbstractRelation.get_relation_classes_of_entity_name(
            entity_name=entity)
        side_bar = []
        for rel in relations:
            match = [
                rel.get_related_entity_classA().__name__.lower(),
                rel.get_related_entity_classB().__name__.lower()
            ]
            prefix = "{}{}-".format(match[0].title()[:2], match[1].title()[:2])
            table = get_generic_relations_table(
                relation_class=rel, entity_instance=instance, detail=False)
            title_card = ''
            if match[0] == match[1]:
                title_card = entity.title()
                dict_1 = {'related_' + entity.lower() + 'A': instance}
                dict_2 = {'related_' + entity.lower() + 'B': instance}
                if 'apis_highlighter' in settings.INSTALLED_APPS:
                    objects = rel.objects.filter_ann_proj(request=request).filter(
                        Q(**dict_1) | Q(**dict_2))
                else:
                    objects = rel.objects.filter(
                        Q(**dict_1) | Q(**dict_2))
            else:
                if match[0].lower() == entity.lower():
                    title_card = match[1].title()
                else:
                    title_card = match[0].title()
                dict_1 = {'related_' + entity.lower(): instance}
                if 'apis_highlighter' in settings.INSTALLED_APPS:
                    objects = rel.objects.filter_ann_proj(
                        request=request).filter(**dict_1)
                else:
                    objects = rel.objects.filter(**dict_1)
            tb_object = table(data=objects, prefix=prefix)
            tb_object_open = request.GET.get(prefix + 'page', None)
            RequestConfig(request, paginate={
                          "per_page": 10}).configure(tb_object)
            side_bar.append((title_card, tb_object, ''.join(
                [x.title() for x in match]), tb_object_open))
        form = get_entities_form(entity.title())
        form = form(instance=instance)
        form_text = FullTextForm(entity=entity.title(), instance=instance)
        if 'apis_highlighter' in settings.INSTALLED_APPS:
            form_ann_agreement = SelectAnnotatorAgreement()
        else:
            form_ann_agreement = False
        if 'apis_bibsonomy' in settings.INSTALLED_APPS:
            apis_bibsonomy = getattr(
                settings, 'APIS_BIBSONOMY_FIELDS', ['self'])
            apis_bibsonomy_texts = getattr(
                settings, "APIS_BIBSONOMY_TEXTS", False)
            if apis_bibsonomy_texts:
                apis_bibsonomy.extend([f"text_{pk}" for pk in TextType.objects.filter(
                    name__in=apis_bibsonomy_texts).values_list('pk', flat=True) if f"text_{pk}" not in apis_bibsonomy])
            if isinstance(apis_bibsonomy, list):
                apis_bibsonomy = '|'.join([x.strip() for x in apis_bibsonomy])
        else:
            apis_bibsonomy = False
        object_revisions = Version.objects.get_for_object(instance)
        object_lod = Uri.objects.filter(entity=instance)
        object_texts, ann_proj_form = get_highlighted_texts(request, instance)
        object_labels = Label.objects.filter(temp_entity=instance)
        tb_label = LabelTableEdit(
            data=object_labels, prefix=entity.title()[:2] + 'L-')
        tb_label_open = request.GET.get('PL-page', None)
        side_bar.append(('Label', tb_label, 'PersonLabel', tb_label_open))
        RequestConfig(request, paginate={"per_page": 10}).configure(tb_label)
        perm = ObjectPermissionChecker(request.user)
        permissions = {'change': perm.has_perm('change_{}'.format(entity), instance),
                       'delete': perm.has_perm('delete_{}'.format(entity), instance),
                       'create': request.user.has_perm('entities.add_{}'.format(entity))}
        template = select_template(["merge/entity_create_generic.html"])

        context = {
            'entity_type': entity,
            'form': form,
            'form_text': form_text,
            'instance': instance,
            'right_card': side_bar,
            'object_revisions': object_revisions,
            'object_texts': object_texts,
            'object_lod': object_lod,
            'ann_proj_form': ann_proj_form,
            'form_ann_agreement': form_ann_agreement,
            'apis_bibsonomy': apis_bibsonomy,
            'permissions': permissions,
            'detail_url': instance.get_absolute_url().replace("apis/entities", "dubletten")
        }
        form_merge_with = GenericEntitiesStanbolForm(entity, ent_merge_pk=pk)
        context['form_merge_with'] = form_merge_with
        if "apis_ampel" in settings.INSTALLED_APPS:
            context["show_ampel"] = is_ampel_active(entity)
        return HttpResponse(template.render(request=request, context=context))

    def post(self, request, *args, **kwargs):
        entity = kwargs['entity']
        pk = kwargs['pk']
        entity_model = AbstractEntity.get_entity_class_of_name(entity)
        instance = get_object_or_404(entity_model, pk=pk)
        form = get_entities_form(entity.title())
        form = form(request.POST, instance=instance)
        form_text = FullTextForm(request.POST, entity=entity.title())
        if form.is_valid() and form_text.is_valid():
            entity_2 = form.save()
            form_text.save(entity_2)
            return redirect(reverse('dubletten_tool:generic_entities_edit_view', kwargs={
                'pk': pk, 'entity': entity
            }))
        else:
            template = select_template(["merge/entity_create_generic.html"])

            perm = ObjectPermissionChecker(request.user)
            permissions = {'change': perm.has_perm('change_{}'.format(entity), instance),
                           'delete': perm.has_perm('delete_{}'.format(entity), instance),
                           'create': request.user.has_perm('entities.add_{}'.format(entity))}
            context = {
                'form': form,
                'entity_type': entity,
                'form_text': form_text,
                'instance': instance,
                'permissions': permissions}
            if entity.lower() != 'place':
                form_merge_with = GenericEntitiesStanbolForm(
                    entity, ent_merge_pk=pk)
                context['form_merge_with'] = form_merge_with
                return TemplateResponse(request, template, context=context)
            return HttpResponse(template.render(request=request, context=context))


@method_decorator(login_required, name='dispatch')
class GenericEntitiesCreateView(View):
    def get(self, request, *args, **kwargs):
        entity = kwargs['entity']
        form = get_entities_form(entity.title())
        form = form()
        form_text = FullTextForm(entity=entity.title())
        permissions = {'create': request.user.has_perm(
            'entities.add_{}'.format(entity))}
        template = select_template(['apis_entities/{}_create_generic.html'.format(entity),
                                    'apis_entities/entity_create_generic.html'])
        return HttpResponse(template.render(request=request, context={
            'entity_type': entity,
            'permissions': permissions,
            'form': form,
            'form_text': form_text}))

    def post(self, request, *args, **kwargs):
        entity = kwargs['entity']
        form = get_entities_form(entity.title())
        form = form(request.POST)
        form_text = FullTextForm(request.POST, entity=entity.title())
        if form.is_valid() and form_text.is_valid():
            entity_2 = form.save()
            form_text.save(entity_2)
            return redirect(reverse('apis:apis_entities:generic_entities_detail_view', kwargs={
                'pk': entity_2.pk, 'entity': entity
            }))
        else:
            permissions = {'create': request.user.has_perm(
                'apis_entities.add_{}'.format(entity))}
            template = select_template(['apis_entities/{}_create_generic.html'.format(entity),
                                        'apis_entities/entity_create_generic.html'])
            return HttpResponse(template.render(request=request, context={
                'permissions': permissions,
                'form': form,
                'form_text': form_text}))


def getRelationsHelperFunction(instance):
    rels = instance.personinstitution_set.values("id", "related_institution", "related_institution__name",
                                                 "relation_type", "relation_type__name", "start_date_written", "end_date_written")
    rels = sorted(rels, key=lambda k: (k["relation_type__name"], k["related_institution__name"], (
        "1400" if not k["start_date_written"] else k["start_date_written"])), reverse=True)
    [el.update({"include": True}) for el in rels]
    return rels


@method_decorator(login_required, name="dispatch")
class GetRelationEditor(TemplateView):
    template_name = "merge/relation_editor.html"

    def get_context_data(self, **kwargs):
        inst = Person.objects.get(id=self.kwargs.get("pk"))
        merged_rel = PersonPersonRelation.objects.get(name="data merged from")
        persons = [r.related_personB for r in PersonPerson.objects.filter(
            related_personA=inst, relation_type=merged_rel)]
        edit_url = inst.get_edit_url().replace("apis/entities", "dubletten")
        detail_url = inst.get_absolute_url().replace("apis/entities", "dubletten")

        rels = getRelationsHelperFunction(inst)
        rels = json.dumps(rels)
        print(rels)
        return {"test": 20, "relations": rels, "detail_url": detail_url, "edit_url": edit_url, "instance": json.dumps({"name": f"{inst.name}, {inst.first_name}", "id": inst.id}), "object": inst, "persons": persons}


@method_decorator(login_required, name="dispatch")
class DeleteRelation(View):

    def dispatch(self, request, **kwargs):
        msg = "success"
        success = True
        try:
            rel_id = kwargs.get("rel_id")
            rel = PersonInstitution.objects.get(id=rel_id)
            rel.delete()
        except Exception as e:
            msg = f"Something went wrong: {e}"
            success = False

        return JsonResponse({"msg": msg, "success": success})


@method_decorator(login_required, name="dispatch")
class CreateRelationView(View):
    # todo: check that this is failsave
    def post(self, request):
        success = True
        msg = True
        re_id = None
        per = get_object_or_404(Person, pk=request.POST.get(
            "id", "-111"))  # todo: check if this is best solution
        data = json.loads(request.POST.get("relation"))
        del data["id"]
        try:
            temp = {}
            temp.update({"related_person": per})
            temp.update({"related_institution": Institution.objects.get(
                id=data["related_institution"])})
            temp.update({"relation_type": PersonInstitutionRelation.objects.get(
                id=data["relation_type"])})
            rel = PersonInstitution.objects.create(**temp)
            rel.save()
        except Exception as e:
            success = False
            msg = f"Couldn't create new Relation:\n {e}"

        return JsonResponse({"success": success, "msg": msg, "rel_id": rel.id})


@method_decorator(login_required, name="dispatch")
class UpdateFieldView(View):
    def post(self, request):
        success = True
        msg = "success"

        data = json.loads(request.POST.get("relation"))
        field_name = request.POST.get("field_name")
        rel = get_object_or_404(PersonInstitution, pk=data["id"])
        try:

            # todo: convert year only to add 06-30 automatically
            if field_name in ["start_date_written", "end_date_written"] and re.match(r"\d{4}", data[field_name].strip()[-4:]) and not "/" in data[field_name]:
                print("date matched", data[field_name])
                data[field_name] = f"{data[field_name]}<{data[field_name].strip()[-4:]}-06-30>"
                print("changed data field name to:", data[field_name])
            elif field_name in ["start_date_written", "end_date_written"] and re.match(r"\d{4}-\d{2}-\d{2}", data[field_name].strip()[-10:]) and not "<" in data[field_name] and not ">" in data[field_name] and not "/" in data[field_name]:
                data[field_name] = f"{data[field_name]}<{data[field_name].strip()[-10:]}>"
                print("caught long date without <")
            else:
                print("date didn't match", data[field_name])
            setattr(rel, field_name, data[field_name])
            rel.save()
            field_value = data[field_name]
        except Exception as e:
            success = False
            msg = f"Couldn't save field:\n{e}"
            field_value = False

        return JsonResponse({"success": success, "msg": msg, "new_value": field_value})



# Generated by Django 3.2 on 2022-10-26 18:49

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('apis_entities', '0004_auto_20200722_1231'),
        ('dubletten_tool', '0002_dublettenlog'),
    ]

    operations = [
        migrations.CreateModel(
            name='Ampel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('note', models.TextField(blank=True, max_length=2000, null=True)),
                ('person', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='apis_entities.person')),
            ],
        ),
    ]





class Merger: 
    """
    HelperClass Storing common logic and functions used in remerge workflows. 
    Allows for creating seperate Workflows for different kind of remerges.
    """

    def __init__(self, *args, **kwargs):
        pass

    


"""
Collect re-usable code snippets / functions in here and import them into our notebooks where needed. 

Please write a short dosctring what the function should do. Once implement, don't change the function without ensuring, that it still works in the context it was written for. 
"""
from django.contrib.contenttypes.models import ContentType
from django.db.models.base import ModelBase
from typing import Callable
from apis_core.apis_entities.models import AbstractEntity
from apis_core.apis_relations.models import AbstractRelation
import pandas as pd



############# Functions to get list of models and list of model names #####


def get_all_relation_classes():
    return AbstractRelation.get_all_relation_classes()

def get_all_entity_classes():
    return AbstractEntity.get_all_entity_classes()

def get_all_entity_names():
    return AbstractEntity.get_all_entity_names()

def get_all_relation_names():
    return AbstractRelation.get_all_relation_names()



############ Functions to get dataframe ##########


def df_from_values(model, values=None):
    """
    Give a model and receive the dataframe constructed from model.objects.values().
    Optional: give a list of field_names, and get only those values
    """
    if not values:
        return pd.DataFrame(model.objects.values())
    else: 
        if isinstance(values, list):
            return pd.DataFrame(model.objects.values(*values))
        else:
            raise TypeError()

############# Functions intendet to be used as decorators #################



def generic_assert_instance_count_integrity(entities:bool=True, types:bool=True, relations:bool=True, relation_types:bool=True, exclude:list=None, include:list=None)->Callable:
    """
    Generic Decorator that accepts arguments to be used in functions that alter data. By default, tests instance count
     Integrity of all 
    entity, relation, type, and relationtype models before and after executing the wrapped function.
    Throws an assertion error if the instance count diverges.
    You can manage the models to be tested in the following ways:
    - on a per group basis: set the model group argument to False.
        example: (entities=False) --> excludes Person, Institution, Work, Event, Place) from the test
    - you can exclude specific models by passing them in a list in the exclude parameter
        example: (exlude=[Person]) NOTE: pass the model-class as an Object, not a string
    - you can include specific models py passing them in the include list
        example: (entities=False, include=[Person, Work]) --> excludes all entites except Person and Work


    # todo: add a return of new errenous instances on check 
    """
    
    ######### check params #######
    for p in [entities, types, relations, relation_types]:
            if not isinstance(p, bool):
                raise TypeError(f"Expected boolean argument, got {p} with type {type(p)} instead.")
    
    if exclude and not isinstance(exclude, list):
        if isinstance(exclude, ModelBase):
            exclude = [exclude]  
        else:
            raise ValueError(f"Param 'exclude' must be an empty list or a list of Model-objects, got {type(exclude)} instead.")
    
    if include and not isinstance(include, list):
        if isinstance(include, ModelBase):
            include = [include]
        else:
            raise ValueError(f"Param 'include' must be an empty list or a list of Model-objects, got {type(exclude)} instead.")
    
    ######## setup data #########
    
    # to check if group is disabled
    arg_mapping = {
        "entities":entities,
        "relations":relations,
        "relation_types":relation_types,
        "types":types
    }
    
    # mapping between the group of models and the lookup dict to be used in destructuring
    model_lookup_dict = {
        "entities": {"app_label":"apis_entities"},
        "types": {"app_label":"apis_vocabularies", "model__icontains":"type"},
        "relations": {"app_label":"apis_relations"},
        "relation_types":{"app_label":"apis_vocabularies", "model__iendswith":"relation"},
    }
    
    def get_model_set(model_lookup_dict:dict)->set:
        
        # to store the references to the model class instances (Objects)
        model_list = []
        
        # populate the model_list
        for key, kwargs in model_lookup_dict.items():

            # check if the group is disabled
            if arg_mapping[key]:
            
                # add all model classes to the model_list
                model_list += [c.model_class() for c in ContentType.objects.filter(**kwargs)]
            

        # exclude unwanted models
        if exclude:
            for ex in exclude:
                if ex in model_list:
                    model_list.remove(ex)

        # include further models or models belonging to a group that has been disabled as a whole
        if include:
            for inc in include:
                if inc not in model_list:
                    model_list.append(inc)

        # deduplicate models
        model_set = set(model_list)

       
        return model_set
    
    model_set = get_model_set(model_lookup_dict)
    
     # setup the model dict that stores the values from before and after the func executes
    count_dict = {m:{"before":-10, "after":-10} for m in model_set}
    
    
    ###### wraper logic ########
    
    def decorator(func:Callable)->Callable:
        def wrapper(*args, **kwargs):
            failed = False 
            err_count = 0         
            # calculate the before val for all models
            for m in count_dict.keys():
                count_dict[m]["before"] = m.objects.count()
            
            # execute the wrapped function and store the return val in res
            res = func(*args,**kwargs)
            
            # calcualte the after val for all models
            for m in count_dict.keys():
                count_dict[m]["after"] = m.objects.count()

            
            # loop through assert - the actual test
            for m, val in count_dict.items():
                before, after = val["before"], val["after"]
                try:
                    assert before == after, f"Failed integrity test for {m};\nInstanceCount diverges:\nBefore: {before}\nAfter: {after}"
                except AssertionError as e:
                    failed=True
                    err_count += 1
                    print(e)
            print(f"Report from generic_assert_instance_count_integrity: tested: {[m.__name__ for m in model_set]}")
            if not failed:
                print("Report from generic_assert_instance_count_integrity: All integrity tests passed.")
            else: 
                print(f"Report from generic_assert_instance_count_integrity: Failed with {err_count} - Errors. Complete report: {count_dict}")
            return res
        return wrapper
    return decorator
    
    

def generic_count_test(test_func=None, model=None):
    """
    Decorator function with arguments to test if a function removes objects in some models.
    Expects either a test_func, that returns the count of objects, like:
    test_func = lambda : len(PersonInstitution.objects.all())
    or a model to construct the test function from.
    """

    # check if both arguments are None and raise error in this case
    if test_func == model == None:
        raise ValueError("Both optional arguments are None, you must either pass in a test_func or a model-object")
    elif test_func == None:
        test_func = lambda: model.objects.count()

    def decorator(func):
        def wrapper(*args, **kwargs):
            # check count before calling decorated function
            value_before = test_func()
            res = func(*args, **kwargs)
            # check count after calling decorated functin
            value_after = test_func()

            # assert nothing was changed
            assert value_before == value_after, f"Count test for {model}, should be the same {value_before} == {value_after}"
            print("test returned:", value_before, value_after)
            return res

        return wrapper

    return decorator


def generic_duplicates_test(model):
    """
    Test if a model has duplicates before and after running a processing function.
    """
    def no_duplicates():
        objects = model.objects.all()
        names = [o.name for o in objects]
        if len(set(names)) == len(names):
            return True
        else:
            return False

    def decorator(func):
        def wrapper(*args, **kwargs):
            assert no_duplicates()
            res = func(*args, **kwargs)
            assert no_duplicates(), f"Duplicate Test for model: {model} should be True (no duplicates), test was {no_duplicates()}"
            return res
        return wrapper
    return decorator




def generic_count_test_all_models():
    rel_models = ContentType.objects.filter(app_name="apis_relations")


from django.test import TestCase



class ErrorLoggerMixinTest(TestCase):

    def setup(self, *args, **kwargs):
        pass
    
    def test_class_setup(self, *args, **kwargs):
        pass

    def tets_class_reset(self, *args, **kwargs):
        pass

    


import django_tables2 as tables
from django.utils.safestring import mark_safe
from django_tables2.utils import A
from django.db.models import Case, When
from apis_core.apis_relations.models import PersonPerson
from django.conf import settings

from apis_core.apis_entities.models import AbstractEntity
from apis_core.apis_metainfo.tables import (
    generic_order_start_date_written,
    generic_order_end_date_written,
    generic_render_start_date_written,
    generic_render_end_date_written
)


input_form = """
  <input type="checkbox" name="keep" value="{}" title="keep this"/> |
  <input type="checkbox" name="remove" value="{}" title="remove this"/>
"""

empty_text_default = 'There are currently no relations'


class VorfinListTable(tables.Table):

    def render_name(self, record, value):
        if value == "":
            return "(No name provided)"
        else:
            return value
    entity = "Person"
    # reuse the logic for ordering and rendering *_date_written
    # Important: The names of these class variables must correspond to the column field name,
    # e.g. for start_date_written, the methods must be named order_start_date_written and render_start_date_written
    order_start_date_written = generic_order_start_date_written
    order_end_date_written = generic_order_end_date_written
    render_start_date_written = generic_render_start_date_written
    render_end_date_written = generic_render_end_date_written


    name = tables.LinkColumn(
        'dubletten_tool:vorfin_detail', #todo: change to local detail view
        args=[entity.lower(), A('pk')],
        empty_values=[]
    )
    export_formats = [
        'csv',
        'json',
        'xls',
        'xlsx',
    ]
   

    class Meta:
        model = AbstractEntity.get_entity_class_of_name(entity_name="person")
        default_cols = ["name", "first_name", "start_date_written", "end_date_written"]


        fields = default_cols
        attrs = {"class": "table table-hover table-striped table-condensed"}

        # quick ensurance if column is indeed a field of this entity
        for col in default_cols:
            if not hasattr(model, col):
                raise Exception(
                    f"Could not find field in entity:Person\n"
                    f"of column (probably defined in 'table_fields' settings): {col}\n"
                )

    def __init__(self, *args, **kwargs):
        if "apis_ampel" in settings.INSTALLED_APPS:
            from apis_ampel.helper_functions import is_ampel_active
            if is_ampel_active("person"):
                self.base_columns['ampel'] = tables.TemplateColumn(template_name = "ampel/ampel_template_column.html", verbose_name="Ampel")


        super().__init__(*args, **kwargs)



def get_generic_relations_table(relation_class, entity_instance, detail=None):
    """
    Creates a table class according to the relation and entity class given by the parameters.

    :param relation_class: the class where the entity_instance can have instantiated relations to
    :param entity_instance: the entity instance of which related relations and entities are to be displayed
    :param detail: boolean : if this Table is to be displayed in an detail or edit UI
    :return: a django-tables2 Table Class tailored for the respective relation class and entity instance
    """

    # create all variables which save the foreign key fields which are different for each relation class
    entity_class_name = entity_instance.__class__.__name__.lower()
    related_entity_class_name_a = relation_class.get_related_entity_classA().__name__.lower()
    related_entity_class_name_b = relation_class.get_related_entity_classB().__name__.lower()
    related_entity_field_name_a = relation_class.get_related_entity_field_nameA()
    related_entity_field_name_b = relation_class.get_related_entity_field_nameB()

    # find out what other entity class the current entity instance in a given relation class is related to
    # (needed for linkg towards instances of related entities)
    if entity_class_name == related_entity_class_name_a == related_entity_class_name_b:
        other_related_entity_class_name = entity_class_name

    elif entity_class_name == related_entity_class_name_a:
        other_related_entity_class_name = related_entity_class_name_b

    elif entity_class_name == related_entity_class_name_b:
        other_related_entity_class_name = related_entity_class_name_a

    else:
        raise Exception(
            "Did not find the entity instance in the given relation class fields!" +
            "Either a wrong entity instance or wrong relation class was passed to this function."
        )


    class RelationTableBase(tables.Table):
        """
        The base table from which detail or edit tables will inherit from in order to avoid redundant definitions
        """

        # reuse the logic for ordering and rendering *_date_written
        # Important: The names of these class variables must correspond to the column field name,
        # e.g. for start_date_written, the methods must be named order_start_date_written and render_start_date_written
        order_start_date_written = generic_order_start_date_written
        order_end_date_written = generic_order_end_date_written
        render_start_date_written = generic_render_start_date_written
        render_end_date_written = generic_render_end_date_written

        class Meta:
            """
            Meta class needed for django-tables2 plugin.
            """

            empty_text = empty_text_default
            model = relation_class
            print("model is", model)

            # the fields list also serves as the defining order of them, as to avoid duplicated definitions
            fields = [
                'start_date_written',
                'end_date_written',
                'other_relation_type',
                "other_related_entity"
            ]
            # reuse the list for ordering
            sequence = tuple(fields)

            # This attrs dictionary I took over from the tables implementation before. No idea if and where it would be needed.
            attrs = {
                "class": "table table-hover table-striped table-condensed",
                "id": related_entity_class_name_a.title()[:2] + related_entity_class_name_b.title()[:2] + "_conn"
            }

        def render_other_related_entity(self, record, value):
            """
            Custom render_FOO method for related entity linking. Since the 'other_related_entity' is a generated annotation
            on the queryset, it does not return the related instance but only the foreign key as the integer it is.
            Thus fetching the related instance is necessary.

            :param record: The 'row' of a queryset, i.e. an entity instance
            :param value: The current column of the row, i.e. the 'other_related_entity' annotation
            :return: related instance
            """

            if value == record.get_related_entity_instanceA().pk :
                return record.get_related_entity_instanceA()

            elif value == record.get_related_entity_instanceB().pk :
                if relation_class == PersonPerson:
                    return str(record.get_related_entity_instanceB())+f" [{record.get_related_entity_instanceB().id}]"
                else: 
                    return record.get_related_entity_instanceB()

            else:
                raise Exception(
                    "Did not find the entity this relation is supposed to come from!" +
                    "Something must have went wrong when annotating for the related instance."
                )


        def __init__(self, data, *args, **kwargs):

            # annotations for displaying data about the 'other side' of the relation.
            # Both of them ('other_related_entity' and 'other_relation_type') are necessary for displaying relations
            # in context to what entity we are calling this from.
            data = data.annotate(
                # In order to provide the 'other instance' for each instance of a table where this whole logic is called from,
                # the queryset must be annotated accordingly. The following Case searches which of the two related instances
                # of a relation queryset entry is the one corresponding to the current entity instance. When found, take the
                # other related entity (since this is the one we are interested in displaying).
                #
                # The approach of using queryset's annotate method allows for per-instance case decision and thus
                # guarantees that the other related entity is always correctly picked,
                # even in case two entities are of the same class.
                other_related_entity=Case(
                    # **kwargs pattern is needed here as the key-value pairs change with each relation class and entity instance.
                    When(**{
                        related_entity_field_name_a + "__pk": entity_instance.pk,
                        "then": related_entity_field_name_b
                    }),
                    When(**{
                        related_entity_field_name_b + "__pk": entity_instance.pk,
                        "then": related_entity_field_name_a
                    }),
                )
            ).annotate(
                # Get the correct side of the relation type given the current entity instance.
                #
                # The approach of using queryset's annotate method allows for per-instance case decision and thus
                # guarantees that the other related entity is always correctly picked,
                # even in case two entities are of the same class.
                other_relation_type=Case(
                When(**{
                    # A->B relation and current entity instance is A, hence take forward name
                    related_entity_field_name_a + "__pk": entity_instance.pk,
                    "then": "relation_type__name"
                }),
                When(**{
                    # A->B relation and current entity instance is B, hence take reverse name.
                    related_entity_field_name_b + "__pk": entity_instance.pk,
                    "then": "relation_type__name_reverse"
                }),
            )
            )
            for an in data:
                if getattr(an, f"{related_entity_field_name_a}_id") == entity_instance.pk:
                    an.other_relation_type = getattr(an.relation_type, "label")
                else:
                    an.other_relation_type = getattr(an.relation_type, "label_reverse")


            super().__init__(data, *args, **kwargs)


    if detail:

        class RelationTableDetail(RelationTableBase):
            """
            Sublcass inheriting the bulk of logic from parent. This table is used for the 'detail' views.
            """

            def __init__(self, data, *args, **kwargs):
                print("in table detail relationtable, data is:", data)

                # Only addition with respect to parent class is which main url is to be used when clicking on a
                # related entity column.
                self.base_columns["other_related_entity"] = tables.LinkColumn(
                    'apis:apis_entities:generic_entities_detail_view',
                    args=[
                        other_related_entity_class_name,
                        A("other_related_entity")
                    ],
                    verbose_name="Related " + other_related_entity_class_name.title()
                )

                if "apis_ampel" in settings.INSTALLED_APPS:
                    from apis_ampel.helper_functions import is_ampel_active
               
                    if is_ampel_active(relation_class.__name__):
                        self.base_columns['ampel'] = tables.TemplateColumn(template_name = "ampel/edit_inline_table_column.html", verbose_name="Ampel")





                super().__init__(data=data, *args, **kwargs)


        return RelationTableDetail


    else:

        class RelationTableEdit(RelationTableBase):
            """
            Sublcass inheriting the bulk of logic from parent. This table is used for the 'edit' view.
            """

            class Meta(RelationTableBase.Meta):
                """
                Additional Meta fields are necessary for editing functionalities
                """

                # This fields list also defines the order of the elements.
                fields = ["delete"] + RelationTableBase.Meta.fields + ["edit"]

                if 'apis_bibsonomy' in settings.INSTALLED_APPS:
                    fields = ["ref"] + fields

                # again reuse the fields list for ordering
                sequence = tuple(fields)


            def __init__(self, *args, **kwargs):

                # Clicking on a related entity will lead also the edit view of the related entity instance
                self.base_columns["other_related_entity"] = tables.LinkColumn(
                    'apis:apis_entities:generic_entities_edit_view',
                    args=[
                        other_related_entity_class_name, A("other_related_entity")
                    ],
                    verbose_name="Related " + other_related_entity_class_name.title()
                )

                # delete button
                self.base_columns['delete'] = tables.TemplateColumn(
                    template_name='apis_relations/delete_button_generic_ajax_form.html'
                )

                # edit button
                self.base_columns['edit'] = tables.TemplateColumn(
                    template_name='apis_relations/edit_button_generic_ajax_form.html'
                )

                # bibsonomy button
                if 'apis_bibsonomy' in settings.INSTALLED_APPS:
                    self.base_columns['ref'] = tables.TemplateColumn(
                        template_name='apis_relations/references_button_generic_ajax_form.html'
                    )

                super().__init__(*args, **kwargs)


        return RelationTableEdit



from . import defaultdict, ErrorLoggerMixin

class PPHelper(ErrorLoggerMixin):
    """
    Helper class for PPKind logic. 
    Split into two classes as I need static fields on PPKind, which enums don't support. 
    Fields are added to PPHelper after PPkind is created.
    
    method: get_expectation. Takes the classification of a relation and returns the 
    expected output classification to check against.
    """
    
    errors = defaultdict(list)
    enum_errors = defaultdict(list)
    _expectation = {}
    
    @classmethod
    def get_expectation(cls, kind):
        res = cls._expectation.get(kind.name, None)
        if res:
            return res
        else: 
            raise KeyError(f"kind: {kind} not found in dict.keys: {cls._expectation.keys()}")
            #return []
    
   


from . import ErrorLoggerMixin, defaultdict, Person, is_dublette

class PersonHelper(ErrorLoggerMixin):
    """
    Helper class that creates the initial sets of persons (singles, dublettten, vorfins).
    
    Adds logic that supports classification of persons, which is handled in PersonType class.
    """
    _singles = []
    _dubletten = []
    _vorfins = []
    _red_groups = []
    
    errors = defaultdict(list)
    enum_errors = defaultdict(list)
    
    
    #@classmethod
    #def calculate_vorfins(cls):
    #    """
    #    Helper to create vorfin set. Redundant with the logic in update_collections, but kept for now.
    #    """
    #    cls._vorfins = Person.objects.filter(name__icontains="vorfin")

    @classmethod
    def get_red_size(cls):
        """
        returns the count of groups of vorfins with red ampel status. Used
        in logging the iteration count.
        """
        return len(cls._red_groups)
    
    @classmethod
    def update_collections(cls):
        """
        Sorts all Persons into vorfin, dubletten and single lists. 
        Throws an error if there is an intersection between the lists.
        """
        vorfins = Person.objects.filter(name__icontains="vorfin")
        dubletten = Person.objects.exclude(personproxy=None)
        dubletten = filter(is_dublette, list(dubletten))
        dubletten = list(dubletten)
        #singles = [el for el in Person.objects.all() if el not in dubletten and el not in vorfins]

        v_set = set(vorfins)
        d_set = set(dubletten)
        p_set = set(Person.objects.all())

        singles = p_set.difference(v_set.union(d_set))
        cls._vorfins = vorfins
        cls._dubletten = dubletten
        cls._singles = singles
        
        test = len(singles)+len(vorfins)+len(dubletten)
        if not Person.objects.count() == test:
            msg = f"Missmatch in collections in PersonHelper. Expected Persons to be split evenly, but got missmatch.\nPersons: {Person.objects.count()}, set union collections: {test}."
            cls.errors["Expectation Missmatch"].append(msg)
            raise Exception(msg)
        



# from functools import reduce
# import importlib

# import django_filters
# from django.conf import settings

# from apis_core.apis_entities.models import *
# from apis_core.apis_vocabularies.models import RelationBaseClass
# from .models import Ampel



# # The following classes define the filter sets respective to their models.
# # Also by what was enabled in the global settings file (or disabled by not explicitley enabling it).
# # Hence by default all filters of all model's fields (automatically generated) and all filters manually defined below
# # are at first created and then deleted by what was enabled in the settings file
# #
# # There is a few overrides happening here, which are in order:
# # 1.) The filters defined in GenericListFilter
# # 2.) The filters automatically defined in model specific ListFilters (by stating exclude = [] in the Meta class)
# # 3.) The filters manually defined in model specific ListFilters (by manual field definitions in the Filter class itself)
# # If anything is redefined in a step further it overrides the field from the step before
# #
# # Additionally, in the global settings file:
# # The filters defined there can provide a dictionary which can have a "method" or "label" key-value pair
# # where then such a key-value from the settings is overriding the respective key-value of a filter defined in this module
# # (e.g. using a different method)

# # TODO __sresch__ : Turn the logic of returing a filter object into a singleton pattern to avoid redundant instantiations
# # TODO __sresch__ : use the order of list of filter fields in settings


# #######################################################################
# #
# #   Generic super class for sharing filters accross all entities
# #
# #######################################################################

# class GenericListFilter(django_filters.FilterSet):

#     class Meta:
#         model = None # must be overriden with entity specific Filter Classes
#         exclude = getattr(settings, "APIS_RELATIONS_FILTER_EXCLUDE", [])

#     name = django_filters.CharFilter(method="name_label_filter", label="Name or Label")
#     collection = django_filters.ModelMultipleChoiceFilter(queryset=Collection.objects.all())

#     # TODO __sresch__ : look into how the date values can be intercepted so that they can be parsed with the same logic as in edit forms
#     start_date = django_filters.DateFromToRangeFilter()
#     end_date = django_filters.DateFromToRangeFilter()

#     # TODO __sresch__ : look into how to change these into auto-complete fields
#     related_entity_name = django_filters.CharFilter(method="related_entity_name_filter", label="related entity")
#     related_relationtype_name = django_filters.CharFilter(method="related_relationtype_name_filter", label="relationtype")


#     def __init__(self, *args, **kwargs):

#         # call super init foremost to create dictionary of filters which will be processed further below
#         super().__init__(*args, **kwargs)

#         def eliminate_unused_filters(default_filter_dict):
#             """
#             Method to read in from the settings file which filters should be enabled / disabled and if there are
#             methods or labels to override the default ones.

#             :param default_filter_dict: the default filter dictionary created on filter class instantiation
#                 (which comprises filters defined: in GenericListFilter, in specific model ListFilter and their defaults)

#             :return: a new dictionary which is a subset of the input dictionary and only contains the filters which
#                 are referenced in the settings file (and if there were methods or labels also referenced, using them)
#             """

#             enabled_filters = settings.APIS_ENTITIES[self.Meta.model.__name__]["list_filters"] 
#             enabled_filters = ["ampel"] + enabled_filters
#             print("enabled_filters", enabled_filters)
#             filter_dict_tmp = {}

#             for enabled_filter in enabled_filters:

#                 if type(enabled_filter) == str and enabled_filter in default_filter_dict:
#                     # If string then just use it, if a filter with such a name is already defined

#                     filter_dict_tmp[enabled_filter] = default_filter_dict[enabled_filter]


#                 elif type(enabled_filter) == dict:
#                     # if a dictionary, then look further into if there is a method or label which overrides the defaults

#                     enabled_filter_key = list(enabled_filter.keys())[0]

#                     if enabled_filter_key in default_filter_dict:

#                         # get the dictionary which contains potential method or label overrides
#                         enabled_filter_settings_dict = enabled_filter[enabled_filter_key]

#                         if "method" in enabled_filter_settings_dict:
#                             default_filter_dict[enabled_filter_key].method = enabled_filter_settings_dict["method"]

#                         if "label" in enabled_filter_settings_dict:
#                             default_filter_dict[enabled_filter_key].label = enabled_filter_settings_dict["label"]

#                         filter_dict_tmp[enabled_filter_key] = default_filter_dict[enabled_filter_key]

#                 else:
#                     raise ValueError("Expected either str or dict as type for an individual filter in the settings file.",
#                             "\nGot instead:", type(enabled_filter))

#             return filter_dict_tmp

#         self.filters = eliminate_unused_filters(self.filters)



#     def construct_lookup(self, value):
#         """
#         Parses user input for wildcards and returns a tuple containing the interpreted django lookup string and the trimmed value
#         E.g.
#             'example' -> ('__icontains', 'example')
#             '*example' -> ('__iendswith', 'example')
#             'example*' -> ('__istartswith', 'example')
#             '"example"' -> ('__iexact', 'example')

#         :param value : str : text to be parsed for *
#         :return: (lookup : str, value : str)
#         """

#         if value.startswith("*") and not value.endswith("*"):

#             value = value[1:]
#             return "__iendswith", value

#         elif not value.startswith("*") and value.endswith("*"):

#             value = value[:-1]
#             return "__istartswith", value

#         elif value.startswith('"') and value.endswith('"'):

#             value = value[1:-1]
#             return "__iexact", value

#         else:

#             if value.startswith("*") and value.endswith("*"):

#                 value = value[1:-1]

#             return "__icontains", value



#     def string_wildcard_filter(self, queryset, name, value):
#         lookup, value = self.construct_lookup(value)
#         return queryset.filter(**{name + lookup : value})



#     def name_label_filter(self, queryset, name, value):
#         # TODO __sresch__ : include alternative names queries

#         lookup, value = self.construct_lookup(value)

#         queryset_related_label=queryset.filter(**{"label__label"+lookup : value})
#         queryset_self_name=queryset.filter(**{name+lookup : value})

#         return ( queryset_related_label | queryset_self_name ).distinct().all()



#     def related_entity_name_filter(self, queryset, name, value):
#         """
#         Searches through the all name fields of all related entities of a given queryset

#         For performance reasons it was not sensible to use django's usual lookup foreign key fields
#         (e.g. Person.objects.filter(personwork__related_work__name__icontains=value))
#         Because such naive (but convenient) django ORM approach starts from a given model, goes to the related relation,
#         goes to the related entity and searches therein. But such lookups create a lot of joins in the underlying sql,
#         which get expensive very fast.

#         So the following algorithms reverses this order and always selects primary keys which are then used further to look
#         for related models.
#         Basically what the following does is:
#         step 1: looks up the search value in tempentity class (since all entities inherit from there) and selects their primary keys
#         step 2: uses the keys from step 1 to filter for related relations and selects their primary keys
#         step 3: uses the keys from step 2 to filter for related entities
#         step 4: merge the results from step 3 into one queryset and returns this.

#         :param queryset: the queryset currently to be filtered on
#         :param name: Here not used, because this method filters on names of related entities
#         :param value: The value to be filtered for, input by the user or a programmatic call
#         :return: filtered queryset
#         """

#         # step 1
#         # first find all tempentities where the lookup and value applies, select only their primary keys.
#         lookup, value = self.construct_lookup(value)
#         tempentity_hit = TempEntityClass.objects.filter(**{"name" + lookup: value}).values_list("pk", flat=True)

#         # list for querysets where relations will be saved which are related to the primary keys of the tempentity list above
#         related_relations_to_hit_list = []

#         # step 2
#         # iterate over every relation related to the current queryset model (e.g for Person: PersonWork, PersonPerson, etc)
#         for relation_class in queryset.model.get_related_relation_classes():

#             # get the related classes and lookup names of the current relation
#             # (e.g. for PersonWork: class Person, class Work, "related_person", "related_work")
#             related_entity_classA = relation_class.get_related_entity_classA()
#             related_entity_classB = relation_class.get_related_entity_classB()
#             related_entity_field_nameA = relation_class.get_related_entity_field_nameA()
#             related_entity_field_nameB = relation_class.get_related_entity_field_nameB()

#             # Within a relation class, there is two fields which relate to entities, check now which of the two
#             # is the same as the current one.
#             if queryset.model == related_entity_classA:

#                 # append filtered relation queryset to list for use later
#                 related_relations_to_hit_list.append(
#                     # filter the current relation for if the other related entity's pk is in the hit list of tempentity class
#                     # from before (e.g. if a tempentity's name contains "seu", then its primary key will be used here, to check
#                     # if there is a related entity which has the same primary key.
#                     relation_class.objects.filter(
#                         **{related_entity_field_nameB + "_id__in": tempentity_hit}
#                     ).values_list(related_entity_field_nameA + "_id", flat=True)
#                 )

#             # also check the related entity field B (because it can be that both A and B are of the same entity class, e.g. PersonPerson,
#             # or that only A is the other one, e.g. when filtering for Work within PersonWork, then the other related class is Person)
#             if queryset.model == related_entity_classB:

#                 # same procedure as above, just with related class and related name being reversed.
#                 related_relations_to_hit_list.append(
#                     relation_class.objects.filter(
#                         **{related_entity_field_nameA + "_id__in": tempentity_hit}
#                     ).values_list(related_entity_field_nameB + "_id", flat=True)
#                 )

#             # A safety check which should never arise. But if it does, we know there is something wrong with the automated wiring of
#             # entities and their respective relation classes (e.g. for Person, there should never be EventWork arising here)
#             if queryset.model != related_entity_classA and queryset.model != related_entity_classB:

#                 raise ValueError("queryset model class has a wrong relation class associated!")

#         # step 3
#         # Filter the entity queryset for each of the resulting relation querysets produced in the loop before
#         queryset_filtered_list = [ queryset.filter(pk__in=related_relation) for related_relation in related_relations_to_hit_list ]

#         # step 4
#         # merge them all
#         result = reduce( lambda a,b : a | b, queryset_filtered_list).distinct()

#         return result




#     def related_relationtype_name_filter(self, queryset, name, value):
#         """
#         Searches through the all name fields of all related relationtypes of a given queryset

#         The following logic is almost identical to the one in method 'related_entity_name_filter', so please look up its
#         comments for documentational purpose therein.

#         Differences are commented however.

#         :param queryset: the queryset currently to be filtered on
#         :param name: Here not used, because this method filters on names of related relationtypes
#         :param value: The value to be filtered for, input by the user or a programmatic call
#         :return: filtered queryset
#         """

#         lookup, value = self.construct_lookup(value)

#         # look up through name and name_reverse of RelationBaseClass
#         relationbaseclass_hit = (
#             RelationBaseClass.objects.filter(**{"name" + lookup: value}).values_list("pk", flat=True) |
#             RelationBaseClass.objects.filter(**{"name_reverse" + lookup: value}).values_list("pk", flat=True)
#         ).distinct()

#         related_relations_to_hit_list = []

#         for relation_class in queryset.model.get_related_relation_classes():

#             related_entity_classA = relation_class.get_related_entity_classA()
#             related_entity_classB = relation_class.get_related_entity_classB()
#             related_entity_field_nameA = relation_class.get_related_entity_field_nameA()
#             related_entity_field_nameB = relation_class.get_related_entity_field_nameB()

#             if queryset.model == related_entity_classA:

#                 # Only difference to method 'related_entity_name_filter' is that the lookup is done on 'relation_type_id'
#                 related_relations_to_hit_list.append(
#                     relation_class.objects.filter(
#                         **{"relation_type_id__in": relationbaseclass_hit}
#                     ).values_list(related_entity_field_nameA + "_id", flat=True)
#                 )

#             if queryset.model == related_entity_classB:

#                 # Only difference to method 'related_entity_name_filter' is that the lookup is done on 'relation_type_id'
#                 related_relations_to_hit_list.append(
#                     relation_class.objects.filter(
#                         **{"relation_type_id__in": relationbaseclass_hit}
#                     ).values_list(related_entity_field_nameB + "_id", flat=True)
#                 )

#             if queryset.model != related_entity_classA and queryset.model != related_entity_classB:

#                 raise ValueError("queryset model class has a wrong relation class associated!")

#         queryset_filtered_list = [ queryset.filter(pk__in=related_relation) for related_relation in related_relations_to_hit_list ]

#         result = reduce( lambda a,b : a | b, queryset_filtered_list).distinct()

#         return result


#     def related_arbitrary_model_name(self, queryset, name, value):
#         """
#         Searches through an arbitrarily related model on its name field.

#         Note that this works only if
#             * the related model has a field 'name'
#             * the filter using this method has the same name as the field of the model on which the filter is applied.
#                 (E.g. the field 'profession' on a person relates to another model: the professiontype. Here the filter on a person
#                 must also be called 'profession' as the field 'profession' exists within the person model and is then used to search in.
#                 Using this example of professions, such a lookup would be generated: Person.objects.filter(profession__name__... ) )
#         """

#         lookup, value = self.construct_lookup(value)

#         # name variable is the name of the filter and needs the corresponding field within the model
#         return queryset.filter( **{ name + "__name" + lookup : value } )



# #######################################################################
# #
# #   Overriding Entity filter classes
# #
# #######################################################################


# class PersonListFilter(GenericListFilter):
#     print("person list filter called")

#     class Meta(GenericListFilter.Meta):
#         model = Person

#     gender = django_filters.ChoiceFilter(choices=(('', 'any'), ('male', 'male'), ('female', 'female')))
#     profession = django_filters.CharFilter(method="related_arbitrary_model_name")
#     title = django_filters.CharFilter(method="related_arbitrary_model_name")
#     name = django_filters.CharFilter(method="person_name_filter", label="Name or Label of person")
#     ampel = django_filters.ChoiceFilter(choices=Ampel.ampel_choices, field_name="ampel__status", label="Ampel")

#     def person_name_filter(self, queryset, name, value):

#         lookup, value = self.construct_lookup(value)

#         queryset_related_label=queryset.filter(**{"label__label"+lookup : value})
#         queryset_self_name=queryset.filter(**{name+lookup : value})
#         queryset_first_name=queryset.filter(**{"first_name"+lookup : value})

#         # return QuerySet.union(queryset_related_label, queryset_self_name, queryset_first_name)
#         return (queryset_related_label | queryset_self_name | queryset_first_name).distinct().all()

#     # def ampel_filter(self, queryset, value, *args, **kwargs):
#     #     print("in ampel filter")
#     #     print(value)
#     #     print(queryset)
#     #     return queryset

# Person.set_list_filter_class(PersonListFilter)


# class PlaceListFilter(GenericListFilter):

#     class Meta(GenericListFilter.Meta):
#         model = Place

#     # TODO __sresch__ : decide on margin tolerance of input, for now the number must be precise
#     lng = django_filters.NumberFilter(label='Longitude')
#     lat = django_filters.NumberFilter(label='Latitude')

# Place.set_list_filter_class(PlaceListFilter)


# class InstitutionListFilter(GenericListFilter):

#     class Meta(GenericListFilter.Meta):
#         model = Institution

# Institution.set_list_filter_class(InstitutionListFilter)


# class EventListFilter(GenericListFilter):

#     class Meta(GenericListFilter.Meta):
#         model = Event

# Event.set_list_filter_class(EventListFilter)


# class WorkListFilter(GenericListFilter):

#     class Meta(GenericListFilter.Meta):
#         model = Work

#     kind = django_filters.ModelChoiceFilter(queryset=WorkType.objects.all())

# Work.set_list_filter_class(WorkListFilter)


# a_ents = getattr(settings, 'APIS_ADDITIONAL_ENTITIES', False)

# if a_ents:
#     with open(a_ents, 'r') as ents_file:
#         ents = yaml.load(ents_file, Loader=yaml.CLoader)
#         for ent in ents['entities']:
#             ent_class = getattr(importlib.import_module(f"apis_core.apis_entities.models"), ent['name'])
#             vocabs = []
#             for v in ent.get('vocabs', []):
#                 ent_class_type = getattr(importlib.import_module(f"apis_core.apis_vocabularies.models"), v)
#                 vocabs.append((v, ent_class_type))
#             class filterMeta:
#                 model = ent_class
#                 exclude = GenericListFilter.fields_to_exclude
#             dict_fc = {
#                 "Meta": filterMeta,
#             }
#             for v in vocabs:
#                 dict_fc[v] =  django_filters.ModelChoiceFilter(queryset=ent_class_type.objects.all())
#             filterclass_ent = type(f"{ent['name'].title()}Filter", (GenericListFilter, ), dict_fc)
#             globals()[f"{ent['name'].title()}ListFilter"] = filterclass_ent


# def get_list_filter_of_entity(entity):
#     """
#     Main method to be called somewhere else in the codebase in order to get the FilterClass respective to the entity string input

#     :param entity: str: type of entity
#     :return: Entity specific FilterClass
#     """
#     if entity == "Person" or entity == "person":
#         print("intercepted filter call")
#         return PersonListFilter
#     print("call went on")
#     entity_class = AbstractEntity.get_entity_class_of_name(entity)
#     filter_class = entity_class.list_filter_class

#     if filter_class is not None:
#         return filter_class
#     else:
#         el = entity.title()
#         try:
#             return globals()[f"{el}ListFilter"]
#         except KeyError:
#             raise ValueError("Could not find respective filter for given entity type:", el)


from . import ErrorLoggerMixin, rt_vorfin, ProcessingTracker, PersonPerson, PersonPersonRelation, PersonProxy, defaultdict, log, PPKind, PPHelper


class RelationWriter(ErrorLoggerMixin):
  

    @classmethod
    def setup(cls):
        cls.errors = defaultdict(list)
        cls.enum_errors = defaultdict(list)


    @classmethod 
    def reset(cls):
        cls.setup()
        
    @classmethod
    def write_person_person_rels(cls, g):
        """
        Updated Version of the write person person rels logic. 

        Now properly takes care of the different types of persons that might be 
        participating in a PersonPerson relation.

        Main logic added is: 

        - it classifies each relation before it is written to the vorfin entry
        - it looks up and stores the expected output for each classified relation 
        - it then classifies the newly created relation and checks against the expectation

        Everything is logged to the logfile. Missmatches are visible by looking for 
        "expectation missmatch" in the logfile.

        """

        model = PersonPerson
        vorfin = g.vorfin  # Person.objects.get(id=MergeGroup.group_map[g.id])
       

        def get_dublette_or_vorfin(per):
            # called in:
            # relation writer
            res = None
            if PersonProxy.objects.filter(person=per).exists():
                proxy = PersonProxy.objects.get(person=per)

                if proxy.status == "single":
                    ##### CountPerPer.single_count += 1
                    res = per
                else:
                    gs = proxy.group_set.all()
                    if len(gs) >= 1:

                        res = gs[
                            0
                        ].vorfin  # Person.objects.get(id=MergeGroup.group_map[gs[0].id])

                    if not res:
                        cls.errors["PersonProxy lookup error in get_dublette_or_vorfin"].append(
                            f"per {per} had proxy {proxy} and was not single {proxy.status}. Group set was: {qs}.")
                        res = per

                    else:
                        # CountPerPer.member_count += 1 # Not sure what this does. Investigate and refactor.
                        pass
            else:
                res = per
            return res

        log(f"in write per per rels --> group is: {g}")
        for m in g.members.all():
            ppA = PersonPerson.objects.filter(
                related_personA=m.person).exclude(relation_type=rt_vorfin)
            ppB = PersonPerson.objects.filter(
                related_personB=m.person).exclude(relation_type=rt_vorfin)
            log(f"member is: {m} ({m.id})")
            if ppA:
                for el in ppA:
                    if el.relation_type != rt_vorfin:
                        type_of_rel_before = PPKind.classify(
                            el, in_step="PPA before")
                        ProcessingTracker.track(el, type_of_rel_before)
                        expectation = PPHelper.get_expectation(
                            type_of_rel_before)
                        log(
                            f"relation (A) of member before : {type_of_rel_before}, rel_id: {el.id}")
                        perB = get_dublette_or_vorfin(el.related_personB)
                        log(f"perB is: {perB}")
                        log(f"related_personA is {vorfin}")
                        temp = {
                            # INSERT EXPECTED RELATION OUTCOME HERE AND TEST AGAINST IT
                            "relation_type": el.relation_type,
                            "start_date_written": el.start_date_written,
                            "end_date_written": el.end_date_written,
                            "related_personA": vorfin,
                            "related_personB": perB,
                        }
                        res, c = model.objects.get_or_create(**temp)
                        type_of_rel_after = PPKind.classify(
                            res, in_step="PPA after")
                        ProcessingTracker.add(el, res, type_of_rel_after)
                        if not type_of_rel_after in expectation:
                            msg = f"Expectation missmatch; expected {expectation}, got {type_of_rel_after} instead. before: {type_of_rel_before}"
                            log(msg)
                            cls.errors["Expectation missmatch"].append(msg)
                        log(
                            f"new A - relation created: {type_of_rel_after}, rel_id: {res.id}, created: {c}")
            if ppB:
                for el in ppB:
                    if el.relation_type != rt_vorfin:
                        type_of_rel_before = PPKind.classify(
                            el, in_step="PPB before")
                        expectation = PPHelper.get_expectation(
                            type_of_rel_before)
                        ProcessingTracker.track(el, type_of_rel_before)
                        log(
                            f"relation (B) of member before : {type_of_rel_before}, rel_id: {el.id}")
                        perA = get_dublette_or_vorfin(el.related_personA)
                        log(f"perA is: {perA}")
                        log(f"realted_personB is: {vorfin}")
                        temp = {
                            "relation_type": el.relation_type,
                            "start_date_written": el.start_date_written,
                            "end_date_written": el.end_date_written,
                            "related_personA": perA,
                            "related_personB": vorfin,
                        }
                        res, c = model.objects.get_or_create(**temp)
                        type_of_rel_after = PPKind.classify(
                            res, in_step="PPB after")
                        ProcessingTracker.add(el, res, type_of_rel_after)
                        if not type_of_rel_after in expectation:
                            msg = f"Expectation missmatch; expected {expectation}, got {type_of_rel_after} instead. before: {type_of_rel_before}"
                            log(msg)
                            cls.errors["Expectation missmatch"].append(msg)
                        log(
                            f"new B - relation created: {type_of_rel_after}, rel_id: {res.id}, created: {c}")



# This init file takes care of shared dependencies in all classes,
# it also takes care of interdependent classes by resolving the import order
# and it also takes care of injecting dependent classes into other classes

# shared imports
from collections import defaultdict, Counter
from enum import Enum
from apis_core.apis_entities.models import Person, AbstractEntity
from apis_core.apis_relations.models import AbstractRelation, PersonPerson
from apis_core.apis_vocabularies.models import PersonPersonRelation, LabelType
from apis_core.apis_metainfo.models import Collection, Label
from dubletten_tool.models import Group, PersonProxy
from apis_ampel.models import AmpelTemp
import pandas as pd

from time import time

rt_vorfin, c = PersonPersonRelation.objects.get_or_create(
            name="data merged from", name_reverse="merged into"
        )
# this is just a dummy to resolve imports
log = print
class LoggerDummy:
     exception = print

logger = LoggerDummy


# helper functions
from dubletten_tool.src.functions.helper_functions import get_group, is_dublette, get_ampel, function_timer
# ordered imports of dependent classes
from .ErrorLoggerMixin import ErrorLoggerMixin

# Helper must be imported before Type
from .PersonHelper import PersonHelper
from .PersonType import PersonType

# set enum in PersonHelper for error logging:
PersonHelper.enum = PersonType


# Helper must be imported before Kind
from .PPHelper import PPHelper
from .PPKind import PPKind


# This static fields in PPHelper use the PPKind Enum above, so they need to be added after the class is initialised
# Mapping between the PersonType of A- and B- side of PP -rels and their PPKind formal representation.
# TODO: does not include all possible connections and outcomes
PPHelper._kindmap = {
    (PersonType.DUBLETTE, PersonType.SINGLE): PPKind.DUBSING,
    (PersonType.SINGLE, PersonType.DUBLETTE): PPKind.SINGDUB,
    (PersonType.VORFIN, PersonType.VORFIN): PPKind.VORVOR,
    (PersonType.SINGLE, PersonType.VORFIN): PPKind.SINGVOR,
    (PersonType.VORFIN, PersonType.SINGLE): PPKind.VORSING,
    (PersonType.SINGLE, PersonType.SINGLE): PPKind.SINGSING,
    (PersonType.DUBLETTE, PersonType.VORFIN): PPKind.DUBVOR,
    (PersonType.VORFIN, PersonType.DUBLETTE): PPKind.VORDUB,
    # VORSAME not included in mapping as the mapping is only used in creating new relations for groups
    # that are to be remerged - i.e. no vorfin entries exist there.

}

# Mapping between initial kind of relation and the expected 'correct' kind of created relations from them.
# Used to check against and raise errors if expectation is not matched during writing of the relations.
PPHelper._expectation = {
    PPKind.DUBSING.name: [PPKind.VORSING],
    PPKind.SINGDUB.name: [PPKind.SINGVOR],
    PPKind.VORVOR.name: [PPKind.VORVOR],
    PPKind.SINGVOR.name: [PPKind.SINGVOR],
    PPKind.VORSING.name: [PPKind.VORSING],
    PPKind.SINGSING.name: [PPKind.SINGSING],
    PPKind.DUBVOR.name: [PPKind.VORVOR],
    PPKind.VORDUB.name: [PPKind.VORVOR],
    PPKind.DUBDUB.name: [PPKind.VORVOR, PPKind.VORSAME],
    PPKind.DUBSAME.name: [PPKind.VORSAME],
}


# set enum in PPHelper for error logging:
PPHelper.enum = PPKind



from .ProcessingTracker import ProcessingTracker
from .RelationWriter import RelationWriter
from .TempRel import TempRel
from .MergeGroup import MergeGroup



from django.urls import path
from django.conf.urls import url
from .views import get_all_notes, getToolPage, getNoteJSON, HandleNoteForm, get_group_suggestions, get_group_ajax, rename_group, update_group_checkboxes, create_new_group, merge_groups, remove_member, get_singles, get_groups, get_single_ajax, get_person_suggestions
from .detail_view import OverwrittenGenericEntitiesDetailView
from .merge_views import DeleteRelation,  GenericEntitiesEditView, GenericEntitiesDetailView, GenericListViewNew, GetRelationEditor, UpdateFieldView, CreateRelationView, RemergeGroupView, UpdateCollectionsView
app_name = "dubletten_tool"

urlpatterns = [
    url(r"entity/update_collections", UpdateCollectionsView.as_view(), name="update_collections"),

    # handles field updates in tables in relation_editor
    url(r"entity/update_field", UpdateFieldView.as_view(), name="update_field"),
    # Handles re run of merge in dubletten_tool
    url(r"run_remerge/(?P<g_id>[0-9]+)/", RemergeGroupView.as_view(), name="remerge_group"),
    # handles creating new relation in dubletten view in relation_editor
    url(r"entity/create_relation", CreateRelationView.as_view(), name="create_relation"),
    # handles deleting a relation in relation_editor
    url(r"entity/(?P<rel_id>[0-9]+)/delete$", DeleteRelation.as_view(), name="delete_relation" ),
    # Edit view of vorfin einträge in edit_merges
    url(
        r"^entity/(?P<entity>[a-z]+)/(?P<pk>[0-9]+)/edit$",
        GenericEntitiesEditView.as_view(),
        name="generic_entities_edit_view",
    ),
    # detail view of vorfin einträge in edit_merges
    url(
        r"^entity/(?P<entity>[a-z]+)/(?P<pk>[0-9]+)/detail$",
        GenericEntitiesDetailView.as_view(),
        name="vorfin_detail",
    ),

    # listview of vorfinale einträge in edit_merges / also the entry point for the edit_merges navlink
    url(
        r"^merge_tool/$",
        GenericListViewNew.as_view(),
        name="tool_merge",
    ),
    # entry point for the relation_editor of the edit_merges page section
     url(
        r"^relation_editor/(?P<pk>[0-9]+)/$",
        GetRelationEditor.as_view(),
        name="relation_editor",
    ),
    path("start/", getToolPage.as_view(), name="tool_page"),
    path("create_group/", create_new_group, name="create_new_group"),
    url(r"^handle_note_form/(?P<g_id>[0-9]+)/(?P<type>[a-z]+)/$", HandleNoteForm.as_view(), name="handle_note_form"),
    path("get_singles/", get_singles, name="get_singles"),
    path("get_all_notes/", get_all_notes, name="get_all_notes"),
    url(r"^get_single_ajax/(?P<s_id>[0-9]+)/$", get_single_ajax, name="get_single_ajax"),
    url(r"^get_singles/(?P<val_name>[a-zA-Zäöüß#\-\_\.()]+)/(?P<val_first>[a-zA-Zäöüß#\_\-\.]+)/(?P<gender>[A-Za-z]+)/$", get_singles, name="get_singles"),
    url(r"^get_groups/(?P<val>[a-zA-Zäüöß_,\[\]\(\)\.]+)/(?P<gender>[A-Za-z]+)/$", get_groups, name="get_groups"),
    url(r"get_note_json/(?P<inst_id>[0-9]+)/(?P<type>[a-z]+)/$", getNoteJSON, name="get_note_json"),
    path("merge_groups/", merge_groups, name="merge_groups"),
    path("rename_group/", rename_group, name="rename_group"),
    url(r"^update_group_checkboxes/(?P<btn_id>[0-9]+)/(?P<group_id>[0-9]+)/$", update_group_checkboxes, name="update_group_checkboxes"),
    url(r"^get_group/(?P<g_id>[0-9]+)/$", get_group_ajax, name="get_group_ajax"),
    url(r"^remove_member/(?P<group_id>[0-9]+)/(?P<per_id>[0-9]+)/$", remove_member, name="remove_member"),
    url(r"^get_person_detail/(?P<per_id>[0-9]+)/$", OverwrittenGenericEntitiesDetailView.as_view(), name="get_person_detail"),
    url(r"^get_person_suggestions/(?P<per_id>[0-9]+)/$", get_person_suggestions, name="get_person_suggestions"),
    url(r"^get_group_suggestions/(?P<group_id>[0-9]+)/$", get_group_suggestions, name="get_group_suggestions"),

]



from dubletten_tool.models import PersonProxy, Group
from apis_ampel.models import AmpelTemp
from apis_core.apis_metainfo.models import Collection
from time import time

def get_group(vorfin):
    """
    Tiny helper function to get group from a vorfin entry.
    Returns Group object.
    """
    try:
        res = Group.objects.get(vorfin=vorfin)
        return res
    except Exception as e:
        print(f"Exception in in get_group: {e} for vorfin {vorfin} ({vorfin.id})")
        return None


def is_dublette(p):
    """
    Tiny helper function to check if Person is a "dublette".
    Returns True or False. 
    TODO: could be updated to also check if there is inconsistent data.
    Like: if a person proxy has status candidate, but is not member in a group.
    Or: if a person proxy is member in several groups.
    Or: if a single is member in a group /several groups.
    """
    try:
        pp = PersonProxy.objects.get(person=p)
        if pp.status == "candidate":
            return True
    except:
        return False


def get_ampel(vorfin):
    if AmpelTemp.objects.filter(temp_entity=vorfin):
        ampel = vorfin.ampel.status
    else:
        ampel = "red"
    return ampel


def update_merging_collections():
    """
    This helper function iterates through all person_proxies and orders each person into the singles or dubletten collection.
    This function can be called in the frontend to update those collections after changes to groups - like removing a member, thus changing it's status.
    These changes need to be reflected in the collections.
    """

    print("update collections called")

    Collection.objects.get(name="Dubletten").delete()
    Collection.objects.get(name="Leopolddaten Singles").delete()

    col_hsv = Collection.objects.get(name="Import HSV full 22-6-21")
    col_hzab = Collection.objects.get(name="Import HZAB full 10-3-21")
    col_acc = Collection.objects.get(name="Import ACCESS full 13-10-21")
    col_new, c = Collection.objects.get_or_create(name="Leopolddaten Singles")
    col_dubl, c = Collection.objects.get_or_create(name="Dubletten")

    def check_if_leopold_single(person):
        if person.status == "single":
            c = person.person.collection.all()
            if col_hsv in c or col_hzab in c:
                person.person.collection.add(col_new)
                person.person.save()

    for pp in PersonProxy.objects.all():
        if pp.status == "candidate":
            pp.person.collection.add(col_dubl)
            pp.person.save()
        else:
            check_if_leopold_single(pp)

    print("update finished")


def function_timer(func):
    
    def wrapper(*args, **kwargs):
        before = time()
        res = func(*args, **kwargs)
        after = time()
        print(after-before)
        return res
    
    return wrapper
            


from . import defaultdict, Counter, log

class ProcessingTracker:
    """
    Helper class to 'track' PP - relations before and after writing them to the 'Vorfinale Einträge'.
    
    Differentiates in all method params between a param:'rel' - a PP - relation given in a dublette or in which
    a dublette participates and a param:'res' a relation, created from a 'rel'.
    """
    
    _data = defaultdict(set)
    _handling_counter = {}
    _result_counter = {}
    _data_key_kind_dict = {}
    group_rel_kind_counter = Counter() # unused at the moment
    
    @classmethod
    def track(cls, rel, rel_kind):
        """
        Method to track which relations where created from initial relations. I.e. PP-relations that 
        a dublette participates in. 
        Tracks how often the same relation is accessed in the code (handling_counter) and 
        creates the initial _data - dictionary, whith key: initial rel, value: list of relations created 
        from this realtion. 
        """
        if cls._data.get(rel, None):
            cls._handling_counter[rel] += 1
            log(f"incremented handling counter to {cls._handling_counter[rel]}")

        else: 
            cls._data_key_kind_dict[rel] = rel_kind
            cls._handling_counter[rel] = 0 
            log(f"set handling counter to 0")

    
    @classmethod
    def add(cls, rel, res, res_kind):
        """
        Method to add created relations to the different data dictionaries. 
        param: rel - initial relation and key in all dictionaries. 
        param: res - relation created from 'rel'
        param: res_kind - PPKind of created relation (res).
        """
        cls._data[rel].add((res, res_kind))
        if cls._result_counter.get(rel, None):
            cls._result_counter[rel] +=1
            log(f"incremented result counter to {cls._result_counter[rel]}")

        else:
            log(f"set result counter to 0")
            cls._result_counter[rel] = 0
            
    @classmethod
    def log(cls):
        """
        Method to be called after all PP-rels where written. 
        Implements checks on the stored data and logs a 
        short report of failures to the logger.
        """
        clear_flag = True

        log("\n ##### REPORT from ProcessingTracker _data")
        for k, v in cls._data.items():
            if len(v) > 1:
                clear_flag = False
                log(f"Divergent entries for rel: {k}.")
                for idx, el in enumerate(v):
                    log(f"Entry_{idx}: {el}")
        if clear_flag:
            log("Result: _data showed nothing suspicious. All good.")
        else:
            clear_flag = True
        log("\n ##### REPORT from ProcessingTracker _handling_counter")
        for k, v in cls._handling_counter.items():
            if v > 1:
                clear_flag = False
                log(f"Handling counter was '{v}' for rel: {k}.")
                log(f"Results are: ")
                for idx, el in enumerate(cls._data[k]):
                    log(f"Entry_{idx}: {el}") 
        if clear_flag:
            log("Result: _handling_counter showed nothing suspicious. All good.")
        
                    


from dubletten_tool.src.classes import MergeGroup, function_timer, TempRel, ErrorLoggerMixin, PersonHelper, PPHelper, logger, rt_vorfin, time, ProcessingTracker, RelationWriter


def remerge_single_group(group, vorfins=None):
    """
    Todo: wrap in error handling that resets all classes accordingly (would be a use-case for a context-manager)
    Todo: fetch the re-created and or changed rel data-frames and display them in browser after remerge.
    """

    # setup classes
    TempRel.setup()
    RelationWriter.setup()
    PersonHelper.update_collections()

    if not vorfins:
        old_vorfin = group.vorfin

        # fetch and store all relations from old vorfin
        if old_vorfin:
            all_rels = old_vorfin.get_related_relation_instances()
            for rel in all_rels:
                if rel.relation_type != rt_vorfin:
                    TempRel(rel)

    else: 
        all_rels = []
        for vorfin in vorfins:
            v_rels = vorfin.get_related_relation_instances()
            all_rels += [rel for rel in v_rels if rel.relation_type != rt_vorfin]

        [TempRel(rel) for rel in all_rels]


    # create new vorfin
    
    mg = MergeGroup(group)
    mg.run_process()


    # Update the PersonHelper collections to include new vorfin
    # TODO: old vorfin or vorfins must be deleted, before updating collections
    PersonHelper.update_collections()


    # create perper relations
    try:
        RelationWriter.write_person_person_rels(group)
    except Exception as e:
        logger.exception(e)

    ProcessingTracker.log()

    if not vorfins and old_vorfin:
        # TODO: I think this is in the wrong place. The old vorfins must be deleted before the collections are updated.
    # delete old vorfin
        old_vorfin.delete()
    elif not vorfins and not old_vorfin: 
        print("vorfins was false and no old_vorfin")
    else: 
        print("vorfins true, old vorfin false")
        [v.delete() for v in vorfins]


    # re-write old relations
    # TODO: test what happens on merg of new group, that has no vorfin and no vorfins. IF this can even happen.
    TempRel.re_create_rels()
    TempRel.log_stats_report()
    RelationWriter.log_report()
    RelationWriter.log_details()

    # show which relations where re-added
    # TODO: export logs and xlsx files and offer them to download or display as html table in frontend
    # - use pandas to_html method ---

    changed_vorfins = TempRel.get_changed_edited_vorfins_dataframe()
    created_rels = TempRel.get_created_rels_dataframe()

    # reset all used classes and free the memory
    TempRel.reset()
    RelationWriter.reset()
    # ErrorLoggerMixin needs special treatment here

    # return vorfin
    new_vorfin = mg.person

    print("new_vorfin id ", new_vorfin.id)
    return {"new_vorfin": new_vorfin, "new_vorfin_id": str(new_vorfin.id), "changed_vorfins": changed_vorfins, "created_rels": created_rels}
